'Excel XLS BIFF File Reader/Writer - Shell for NPOI 1.2.2 alpha2
'Written by Hunter Beanland: www.beanland.net.au
'ver 1.1 2010-04-07 +TableToXLS, fix ReadCell exception if value is nothing
'ver 1.2 2010-05-04 Fixed IsNothing(xlRow) in ReadCell/CanReadCell
'ver 1.4 2010-06-25 Update to NPOI1.2.2. Enums, Colour, AutoWidth & Fixes in TableControlsToXLS
Imports System
Imports System.Web
Imports System.Data
Imports System.Collections
Imports Microsoft.VisualBasic
Imports System.IO
Imports NPOI.HSSF.UserModel
Imports NPOI.HPSF
Imports NPOI.POIFS.FileSystem

    'Sample Code:
    'Dim xr As New Excel
    'xr.Open("")
    'While xr.ReadRow
    '    While xr.CanReadCell()
    '        lblMessage.Text &= "," & xr.ReadCell().ToString
    '    End While
    '    lblMessage.Text &= "<p/>"
    'End While
    'xr.WriteNewRow()
    'xr.WriteCell("text")
    'xr.WriteCell(1234)
    'xr.WriteCell(Now)
    'xr.WriteCellAsDate(Now, "dd/MM/yy")
    'xr.FormatLastCell(True, HSSFCellStyle.ALIGN_RIGHT, 15, 2, NPOI.HSSF.Util.HSSFColor.BLUE.index, NPOI.HSSF.Util.HSSFColor.OLIVE_GREEN.index, True)
    'xr.Save("c:\path\sample data_OUT.xls")
    

    ''' <summary>
    ''' Provides easy to use API for basic Excel 97+ BIFF file reading and writing. Exposes the NPOI Excel objects for advanced usage.
    ''' </summary>
    ''' <remarks></remarks>
    Public Class Excel
        Public ErrorMessage As String
        ''' <summary>
        ''' Return String.Empty or Nothing if error (such as no more rows/columns)
        ''' </summary>
        Public ReturnEmptyOnError As Boolean
        Public xlWorkbook As HSSFWorkbook
        Public xlSheet As HSSFSheet
        Public xlRow As HSSFRow
        Public xlCell As HSSFCell
        Public CurrentRow, CurrentColumn As Integer

        ''' <summary>
        ''' When creating this object for running in the normal ASP.NET process, supply no params
        ''' </summary>
        ''' <remarks></remarks>
        Public Sub New()
 
        End Sub


        ''' <summary>
        ''' Open the XLS file for reading
        ''' </summary>
        ''' <param name="XlsFilePath">Path to the file to open. Empty/not specified for new file</param>
        ''' <returns>True if opened successfully</returns>
        ''' <remarks></remarks>
        Public Function Open(Optional ByVal XlsFilePath As String = "") As Boolean
            Try
                If XlsFilePath = String.Empty Then
                    xlWorkbook = New HSSFWorkbook()
                    xlSheet = xlWorkbook.CreateSheet()
                Else
                    Using fs As New FileStream(XlsFilePath, FileMode.Open)
                        xlWorkbook = New HSSFWorkbook(fs)
                        xlSheet = xlWorkbook.GetSheetAt(0)
                        fs.Close()
                    End Using
                End If
                CurrentRow = -1
                CurrentColumn = -1
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function

        ''' <summary>
        ''' See if the next row has data. Return False if end of sheet.
        ''' </summary>
        ''' <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if read ok</returns>
        ''' <remarks></remarks>
        Public Function ReadRow(Optional ByVal Row As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                Else
                    CurrentRow += 1
                End If
                xlRow = xlSheet.GetRow(CurrentRow)
                If IsNothing(xlRow) Then
                    Return False
                Else
                    Return True
                End If
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Tests if we can read the next cell (it exists)
        ''' </summary>
        ''' <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if the cell exists</returns>
        ''' <remarks></remarks>
        Public Function CanReadCell(Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Boolean
            If Row > -1 Then xlRow = xlSheet.GetRow(Row)
            If IsNothing(xlRow) Then Return False
            If Column > -1 Then
                If Column > xlRow.PhysicalNumberOfCells - 1 Then Return False
            Else
                If CurrentColumn + 1 > xlRow.PhysicalNumberOfCells - 1 Then Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell). 
        ''' </summary>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or Nothing/Empty if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCell(Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Object
            If Row > -1 Then
                CurrentRow = Row
                xlRow = xlSheet.GetRow(Row)
            End If
            If IsNothing(xlRow) Then Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
            If Column > -1 Then
                CurrentColumn = Column
            Else
                CurrentColumn += 1
            End If
            If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then
                Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
            End If

            If IsNothing(xlRow.GetCell(CurrentColumn)) Then
                Return String.Empty
            Else
                Select Case xlRow.GetCell(CurrentColumn).CellType
                    Case HSSFCellType.STRING
                        Return xlRow.GetCell(CurrentColumn).StringCellValue
                    Case HSSFCellType.NUMERIC
                        Return xlRow.GetCell(CurrentColumn).NumericCellValue
                    Case HSSFCellType.BOOLEAN
                        Return xlRow.GetCell(CurrentColumn).BooleanCellValue
                    Case Else
                        Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
                End Select
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as Integer type. Strings are attempted to be converted to Integer
        ''' </summary>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or 0 if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCellAsInteger(Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Integer
            Dim int As Object = ReadCell(Row, Column)
            If int.GetType Is GetType(Integer) Then
                Return int
            ElseIf int.GetType Is GetType(String) Then
                Dim xlInteger As Integer
                Integer.TryParse(int, xlInteger)
                Return xlInteger
            Else
                Return 0
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as Double type. Strings are attempted to be converted to Double
        ''' </summary>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or 0 if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCellAsDouble(Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Double
            Dim dbl As Object = ReadCell(Row, Column)
            If dbl.GetType Is GetType(Double) Then
                Return dbl
            ElseIf dbl.GetType Is GetType(String) Then
                Dim xlDouble As Double
                Double.TryParse(dbl, xlDouble)
                Return xlDouble
            Else
                Return 0
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as DateTime type. Strings are attempted to be converted to DateTime
        ''' </summary>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or MinValue if read error</returns>
        ''' <remarks></remarks>   
        Public Function ReadCellAsDate(Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As DateTime
            Dim dbl As Object = ReadCell(Row, Column)
            If dbl.GetType Is GetType(Double) Then
                Return xlRow.GetCell(CurrentColumn).DateCellValue
            ElseIf dbl.GetType Is GetType(String) Then
                Dim xlDate As DateTime
                DateTime.TryParse(dbl, xlDate)
                Return xlDate
            Else
                Return DateTime.MinValue
            End If
        End Function

        ''' <summary>
        ''' Write the value at the next cell (or at the specified cell). 
        ''' </summary>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteCell(ByVal Value As Object, Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                    xlRow = xlSheet.GetRow(Row)
                    If IsNothing(xlRow) Then
                        xlRow = xlSheet.CreateRow(Row)
                        CurrentColumn = -1
                    End If
                Else
                    If IsNothing(xlRow) Then
                        If CurrentRow = -1 Then CurrentRow = 0
                        xlRow = xlSheet.CreateRow(CurrentRow)
                        CurrentColumn = -1
                    End If
                End If
                If Column > -1 Then
                    CurrentColumn = Column
                Else
                    CurrentColumn += 1
                End If
                If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then xlCell = xlRow.CreateCell(CurrentColumn)
                If Value.GetType Is GetType(String) Then
                    xlCell.SetCellType(HSSFCellType.STRING)
                ElseIf Value.GetType Is GetType(Integer) Or GetType(ValueType) Is GetType(Decimal) Or GetType(ValueType) Is GetType(Double) Then
                    xlCell.SetCellType(HSSFCellType.NUMERIC)
                ElseIf Value.GetType Is GetType(DateTime) Then
                    Dim format As HSSFDataFormat = xlWorkbook.CreateDataFormat()
                    Dim dateStyle As HSSFCellStyle = xlWorkbook.CreateCellStyle()
                    dateStyle.DataFormat = format.GetFormat(Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortDatePattern & " HH:mm:ss")
                    xlCell.CellStyle = dateStyle
                ElseIf Value.GetType Is GetType(Boolean) Then
                    xlCell.SetCellType(HSSFCellType.BOOLEAN)
                End If
                xlCell.SetCellValue(Value)
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Write the date/teime value at the next cell (or at the specified cell) with the specified format. 
        ''' </summary>
        ''' <param name="Value">Date/Time value</param>
        ''' <param name="DateTimeFormat">Format to display the date/time in. ie: "dd/MM/yyyy HH:mm:ss AM/PM"</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteCellAsDate(ByVal Value As DateTime, ByVal DateTimeFormat As String, Optional ByVal Row As Integer = -1, Optional ByVal Column As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                    xlRow = xlSheet.GetRow(Row)
                    If IsNothing(xlRow) Then
                        xlRow = xlSheet.CreateRow(Row)
                        CurrentColumn = -1
                    End If
                Else
                    If IsNothing(xlRow) Then
                        If CurrentRow = -1 Then CurrentRow = 0
                        xlRow = xlSheet.CreateRow(CurrentRow)
                        CurrentColumn = -1
                    End If
                End If
                If Column > -1 Then
                    CurrentColumn = Column
                Else
                    CurrentColumn += 1
                End If
                If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then xlCell = xlRow.CreateCell(CurrentColumn)
                If String.IsNullOrEmpty(DateTimeFormat) Then DateTimeFormat = Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortDatePattern & " HH:mm:ss"
                Dim format As HSSFDataFormat = xlWorkbook.CreateDataFormat()
                Dim dateStyle As HSSFCellStyle = xlWorkbook.CreateCellStyle()
                dateStyle.DataFormat = format.GetFormat(DateTimeFormat)
                xlCell.SetCellType(HSSFCellType.NUMERIC)
                xlCell.CellStyle = dateStyle
                xlCell.SetCellValue(Value)
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Set the cell style of the last cell
        ''' </summary>
        ''' <param name="Bold">Bold font. Default = False</param>
        ''' <param name="Align">Cell alignment. Default = CellHorizontalAlignment.GENERAL</param>
        ''' <param name="Width">Cell Width. Default = 10</param>
        ''' <param name="CellSpan">Cell span. Default = 1</param>
        ''' <param name="FGColour">Font colour. Default = NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index</param>
        ''' <param name="BGColour">Background colour. Default = NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index</param>
        ''' <param name="Border">Thin black border. Default = False</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function FormatLastCell(ByVal Bold As Boolean, ByVal Align As CellHorizontalAlignment, ByVal Width As Integer, ByVal CellSpan As Integer, ByVal FGColour As Short, ByVal BGColour As Short, ByVal Border As Boolean) As Boolean
            Dim xlStyle As HSSFCellStyle = xlCell.CellStyle
            If IsNothing(xlStyle) Then xlStyle = xlWorkbook.CreateCellStyle()
            If Bold Or FGColour <> NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index Then
                Dim xlFont As HSSFFont = xlWorkbook.CreateFont()
                If FGColour <> NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index Then xlFont.Color = FGColour
                If Bold Then xlFont.Boldweight = HSSFFont.BOLDWEIGHT_BOLD
                xlStyle.SetFont(xlFont)
            End If
            If BGColour <> NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index Then
                If BGColour <> NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index Then xlStyle.FillForegroundColor = BGColour
                xlStyle.FillPattern = CellFillPattern.SOLID_FOREGROUND
            End If
            If Border Then
                xlStyle.BorderTop = CellBorderType.THIN
                xlStyle.BorderLeft = CellBorderType.THIN
                xlStyle.BorderRight = CellBorderType.THIN
                xlStyle.BorderBottom = CellBorderType.THIN
            End If
            If Align <> CellHorizontalAlignment.GENERAL Then
                xlStyle.Alignment = Align
            End If
            xlCell.CellStyle = xlStyle
            If CellSpan > 1 Then
                xlSheet.AddMergedRegion(New NPOI.HSSF.Util.Region(CurrentRow, CurrentColumn, CurrentRow, CurrentColumn + CellSpan))
            End If
            xlSheet.SetColumnWidth(CurrentColumn, Width * 256)
        End Function

        ''' <summary>
        ''' Creates a new row for writing. Use this if using WriteCell with no row/column specified.
        ''' </summary>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteNewRow() As Boolean
            CurrentRow += 1
            Try
                xlRow = xlSheet.CreateRow(CurrentRow)
                CurrentColumn = -1
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Save the workbook to a new file
        ''' </summary>
        ''' <param name="XlsFilePath">Absolute path and filename for the .xls</param>
        ''' <returns>True if saved</returns>
        ''' <remarks></remarks>
        Public Function Save(ByVal XlsFilePath As String) As Boolean
            Try
                If File.Exists(XlsFilePath) Then File.Delete(XlsFilePath)
                Using fs As New FileStream(XlsFilePath, FileMode.Create)
                    xlWorkbook.Write(fs)
                    fs.Flush()
                    fs.Close()
                End Using
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Save a Table with the formatting (column names from sort order and column header text) from a GridView to an Excel 97+ BIFF .xls file
        ''' </summary>
        ''' <param name="SourceData">Source table</param>
        ''' <param name="GridViewFormat">GridView to read column names and order from</param>
        ''' <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
        ''' <returns>True if successful</returns>
        ''' <remarks></remarks>
        Public Overloads Function TableToXLS(ByRef SourceData As DataTable, ByRef GridViewFormat As GridView, ByVal FileToSave As String) As Boolean
            Dim SourceColOrder(GridViewFormat.Columns.Count - 1) As String
            Dim SourceColText(GridViewFormat.Columns.Count - 1) As String
            Dim intCol As Integer
            For Each dcf As DataControlField In GridViewFormat.Columns
                If dcf.Visible Then
                    SourceColOrder(intCol) = dcf.SortExpression
                    If SourceColOrder(intCol).Contains(".") Then SourceColOrder(intCol) = SourceColOrder(intCol).Substring(SourceColOrder(intCol).IndexOf(".") + 1)
                    SourceColText(intCol) = dcf.HeaderText
                    intCol += 1
                End If
            Next
            ReDim Preserve SourceColOrder(intCol - 1)
            ReDim Preserve SourceColText(intCol - 1)
            Return TableToXLS(SourceData, SourceColOrder, SourceColText, True, FileToSave)
        End Function

        ''' <summary>
        ''' Save a Table to an Excel 97+ BIFF .xls file
        ''' </summary>
        ''' <param name="SourceData">Source table</param>
        ''' <param name="SourceColOrder">Source column name order. Nothing if auto order (per table). Required if SourceColText specified</param>
        ''' <param name="SourceColText">Source column header text. Nothing if auto names (per table). Required if SourceColOrder specified</param>
        ''' <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
        ''' <param name="IncludeHeader">Write column names</param>
        ''' <returns>True if successful</returns>
        ''' <remarks></remarks>
        Public Overloads Function TableToXLS(ByRef SourceData As DataTable, ByRef SourceColOrder() As String, ByRef SourceColText() As String, ByVal IncludeHeader As Boolean, ByVal FileToSave As String) As Boolean
            Try
                If IO.File.Exists(FileToSave) Then IO.File.Delete(FileToSave)
            Catch
                'HttpContext.Current.Trace.Warn("Can not delete previous file")
            End Try
            Dim fiSave As New FileInfo(FileToSave)
            Dim SheetName As String = fiSave.Name.Substring(0, fiSave.Name.IndexOf("."))
            SheetName = Regex.Replace(SheetName, "[^A-Z]", String.Empty, RegexOptions.IgnoreCase)
            Dim blnNoError As Boolean = True

            xlWorkbook = New HSSFWorkbook()

            Dim dsi As DocumentSummaryInformation = PropertySetFactory.CreateDocumentSummaryInformation()
            dsi.Company = "{company name here}"
            xlWorkbook.DocumentSummaryInformation = dsi
            Dim si As SummaryInformation = PropertySetFactory.CreateSummaryInformation()
            si.Title = SheetName
            'si.Author = "{user name here}"
            'si.ApplicationName = "{application's name}"
            xlWorkbook.SummaryInformation = si

            xlSheet = xlWorkbook.CreateSheet(SheetName)
            xlSheet.DisplayGridlines = False
            Dim xlStyle As HSSFCellStyle
            Dim MaxColumns As Integer
            CurrentRow = 0

            'Write header
            If IncludeHeader Then
                xlRow = xlSheet.CreateRow(CurrentRow)
                If IsNothing(SourceColText) Then
                    'Auto names
                    ReDim SourceColOrder(SourceData.Columns.Count - 1)
                    For Each dcCol As DataColumn In SourceData.Columns
                        SourceColOrder(CurrentColumn) = dcCol.ColumnName
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlStyle = xlWorkbook.CreateCellStyle()
                        Dim xlFont As HSSFFont = xlWorkbook.CreateFont()
                        xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index
                        xlFont.Boldweight = HSSFFont.BOLDWEIGHT_BOLD
                        xlStyle.SetFont(xlFont)
                        xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.ROYAL_BLUE.index
                        xlStyle.FillPattern = CellFillPattern.SOLID_FOREGROUND
                        xlCell.SetCellValue(dcCol.ColumnName.Replace("_", " "))
                        xlCell.CellStyle = xlStyle
                        CurrentColumn += 1
                    Next
                Else
                    'Specified names
                    For Each strName As String In SourceColText
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlStyle = xlWorkbook.CreateCellStyle()
                        Dim xlFont As HSSFFont = xlWorkbook.CreateFont()
                        xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index
                        xlFont.Boldweight = HSSFFont.BOLDWEIGHT_BOLD
                        xlStyle.SetFont(xlFont)
                        xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.ROYAL_BLUE.index
                        xlStyle.FillPattern = CellFillPattern.SOLID_FOREGROUND
                        xlCell.SetCellValue(strName)
                        xlCell.CellStyle = xlStyle
                        CurrentColumn += 1
                    Next
                End If
            End If

            If IsNothing(SourceColOrder) Then
                'Get column names if not specified
                CurrentColumn = 0
                ReDim SourceColOrder(SourceData.Columns.Count - 1)
                For Each dcCol As DataColumn In SourceData.Columns
                    SourceColOrder(CurrentColumn) = dcCol.ColumnName
                    CurrentColumn += 1
                Next
            End If

            'Write data rows
            Dim SourceDataCell As Object
            For Each drvRow As DataRow In SourceData.Rows
                CurrentRow += 1
                CurrentColumn = 0
                xlRow = xlSheet.CreateRow(CurrentRow)
                For Each strColName As String In SourceColOrder 'For CurrentColumn = 0 To SourceData.Columns.Count - 1
                    If strColName <> String.Empty Then
                        SourceDataCell = drvRow(strColName)
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlStyle = xlWorkbook.CreateCellStyle()
                        xlStyle.BorderLeft = CellBorderType.THIN
                        xlStyle.BorderRight = CellBorderType.THIN
                        xlStyle.BorderBottom = CellBorderType.THIN
                        xlCell.CellStyle = xlStyle
                        If SourceDataCell.GetType Is GetType(System.Int32) Or SourceDataCell.GetType Is GetType(System.Int64) Then
                            xlCell.SetCellType(NPOI.HSSF.Util.HSSFDataValidation.DATA_TYPE_INTEGER) 'HSSFCellType seems to be missing types
                            If Not IsDBNull(SourceDataCell) Then xlCell.SetCellValue(SourceDataCell)
                        ElseIf SourceDataCell.GetType Is GetType(System.DateTime) Then
                            xlCell.SetCellType(NPOI.HSSF.Util.HSSFDataValidation.DATA_TYPE_DATE)
                            If Not IsDBNull(SourceDataCell) Then xlCell.SetCellValue(SourceDataCell.ToString)
                        ElseIf SourceDataCell.GetType Is GetType(System.Double) Or SourceDataCell.GetType Is GetType(System.Decimal) Then
                            xlCell.SetCellType(NPOI.HSSF.Util.HSSFDataValidation.DATA_TYPE_DECIMAL)
                            If Not IsDBNull(SourceDataCell) Then xlCell.SetCellValue(SourceDataCell)
                        Else
                            xlCell.SetCellValue(SourceDataCell.ToString)
                        End If
                        If MaxColumns < CurrentColumn Then MaxColumns = CurrentColumn
                        CurrentColumn += 1
                    End If
                Next
            Next

            'Set column widths
            For intColumn As Integer = 0 To MaxColumns
                xlSheet.AutoSizeColumn(intColumn)
            Next

            Try
                If FileToSave.IndexOf(":") < 1 Then
                    'just filename to stream to browser and force it to download the file (and launch Excel or save, not display in the browser)
                    Dim ms As New MemoryStream()
                    xlWorkbook.Write(ms)
                    '*** Uncomment if using ASP.NET ***
                    'HttpContext.Current.Response.Clear()
                    'HttpContext.Current.Response.AddHeader("Content-Disposition", "attachment; filename=" & FileToSave)
                    'HttpContext.Current.Response.AddHeader("Content-Length", ms.Length.ToString())
                    'HttpContext.Current.Response.ContentType = "application/octet-stream"
                    'HttpContext.Current.Response.BinaryWrite(ms.GetBuffer())
                    'HttpContext.Current.Response.Flush()
                    'HttpContext.Current.Response.End()
                Else
                    Using fs As New FileStream(FileToSave, FileMode.Create)
                        xlWorkbook.Write(fs)
                        fs.Close()
                    End Using
                End If
            Catch ex As Exception
                blnNoError = False
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
            End Try
            Return blnNoError
        End Function

        Private Function GetXLColour(ByVal SystemColour As System.Drawing.Color) As Short
            'Just map the colours we usually use in our reports to Excel colours - add yours here!
            Select Case SystemColour
                Case Color.White
                    Return NPOI.HSSF.Util.HSSFColor.WHITE.index
                Case Color.DarkBlue
                    Return NPOI.HSSF.Util.HSSFColor.DARK_BLUE.index
                Case Color.Gray
                    Return NPOI.HSSF.Util.HSSFColor.GREY_50_PERCENT.index
                Case Color.LightBlue
                    Return NPOI.HSSF.Util.HSSFColor.PALE_BLUE.index
                Case Color.LightGreen
                    Return NPOI.HSSF.Util.HSSFColor.LIGHT_GREEN.index
                Case Else
                    Return NPOI.HSSF.Util.HSSFColor.AUTOMATIC.index
            End Select
        End Function

    End Class
