﻿//Excel XLS BIFF File Reader/Writer - Shell for NPOI 1.2.4
// ***** This is a generated C# port from a VB class. It may needs some adjustments. Default parameters may require C# 4.0 *****
// ***** Written by Hunter Beanland: www.beanland.net.au
//v1.1 2010-04-07 +TableToXLS, fix ReadCell exception if value is nothing
//v1.2 Fixed IsNothing(xlRow) in ReadCell/CanReadCell
//v1.3 Fixed visible & no sort in TableToXLS
//v1.4 Update to NPOI1.2.2 (new Enums). Colour
//v1.5 Updated to NPOI1.2.3 (NPOI.SS.UserModel). Date/Time data format set properly
//v1.6 Fixed FontBoldWeight
//v1.7 Much better style management
//v2.0 Open and save to streams. Swapped Col and Row parameters. Fixed formula type
//v2.1 GetXLColourBG supports adding custom colours to Excel. GetXLColourFG for font colour - gets nearest colour due to palette bug in 1.2.3. Fixed merged cell border
//v2.2 Fix GetXLColourBG in 1.2.4, Fix setting of date and decimal styles in TableControlsToXLS, GridViewToXLS

using System;
using System.Web;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.IO;
using NPOI.HSSF.UserModel;
using NPOI.SS.UserModel;
using NPOI.HPSF;
using NPOI.POIFS.FileSystem;

//Namespace Bean.Excel

//Sample Code (to write a new file a read it back again):
//private NPOIExcel xr;
//xr.Open();
//xr.WriteNewRow();
//xr.WriteCell("text");
//xr.WriteCell(1234);
//xr.WriteCell(Now);
//xr.WriteCellAsDate(Now, "dd/mm/yy");
//xr.FormatLastCell(True, NPOI.SS.UserModel.HorizontalAlignment.RIGHT, 15, 2, NPOI.HSSF.Util.HSSFColor.BLUE.index, NPOI.HSSF.Util.HSSFColor.OLIVE_GREEN.index, True);
//xr.Save("c:\\web\\sample data.xls");
//'Read it back again
//xr.Open("c:\\web\\sample data.xls");
//while (xr.ReadRow) {
//   while (xr.CanReadCell()) {
//      lblMessage.Text += "," + xr.ReadCell().ToString;
//   }
//   lblMessage.Text += "<p/>";
//}

/// <summary>
/// Provides easy to use API for basic Excel 97+ BIFF file reading and writing. Exposes the NPOI Excel objects for advanced usage.
/// </summary>
/// <remarks></remarks>
public class NPOIExcel
{
	public string ErrorMessage;
	/// <summary>
	/// Return String.Empty or Nothing if error (such as no more rows/columns)
	/// </summary>
	public bool ReturnEmptyOnError;
	public HSSFWorkbook xlWorkbook;
	public HSSFSheet xlSheet;
	public HSSFRow xlRow;
	public HSSFCell xlCell;
	public int CurrentRow;
	public int CurrentColumn;
	public HSSFCellStyle CellStyleDate;
	public HSSFCellStyle CellStyleDateTime;
		//for user defined styles. Reuse existing styles (avoid limit to number of new styles)
	private Generic.Dictionary<string, HSSFCellStyle> CellStyleList = new Generic.Dictionary<string, HSSFCellStyle>();
	private short FormatDate;
	private short FormatDateTime;
	private string DateFormat;
    private string DateTimeFormat;
    
	/// <summary>
	/// Open the XLS file for reading OR create a new xls workbook.
	/// </summary>
	/// <returns>True if opened successfully</returns>
	/// <remarks>Overload added for older c# vers</remarks>
    public bool Open()
    {
        Open(String.Empty);
    }

	/// <summary>
	/// Open the XLS file for reading OR create a new xls workbook. (Does not use file or stream until Save is called)
	/// </summary>
	/// <param name="XlsFilePath">Path to the file to open. Empty/not specified for new file</param>
	/// <returns>True if opened successfully</returns>
	/// <remarks></remarks>
	public bool Open(string XlsFilePath)
	{
		try {
			if (XlsFilePath == string.Empty) {
				xlWorkbook = new HSSFWorkbook();
				CreateStandardProperties("Sheet1");
			} else {
				if (File.Exists(XlsFilePath)) {
					using (FileStream fs = new FileStream(XlsFilePath, FileMode.Open)) {
						xlWorkbook = new HSSFWorkbook(fs);
						if (xlWorkbook.NumberOfSheets > 0) {
							xlSheet = xlWorkbook.GetSheetAt(0);
						} else {
							CreateStandardProperties(XlsFilePath);
						}
						fs.Close();
					}
				} else {
					xlWorkbook = new HSSFWorkbook();
					CreateStandardProperties(XlsFilePath);
				}
			}
			CurrentRow = -1;
			CurrentColumn = -1;
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
			return false;
		}
	}

	/// <summary>
	/// Open the XLS file stream for reading
	/// </summary>
	/// <param name="XlsFileStream">Stream or FileStream to open</param>
	/// <returns>True if opened successfully</returns>
	/// <remarks></remarks>
	public bool Open(Stream XlsFileStream)
	{
		try {
			xlWorkbook = new HSSFWorkbook(XlsFileStream);
			if (XlsFileStream.Length == 0) {
				CreateStandardProperties("Sheet1");
			} else {
				xlSheet = xlWorkbook.GetSheetAt(0);
			}
			CurrentRow = -1;
			CurrentColumn = -1;
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
			return false;
		}
	}

	private void CreateStandardProperties(string SheetName)
	{
		//Create some standard properties and styles for new spreasheets
		xlSheet = xlWorkbook.CreateSheet(SheetName);
		DocumentSummaryInformation dsi = PropertySetFactory.CreateDocumentSummaryInformation();
		//dsi.Company = "{Company Name}"
		xlWorkbook.DocumentSummaryInformation = dsi;
		SummaryInformation si = PropertySetFactory.CreateSummaryInformation();
		si.Title = SheetName;
		//si.Author = "{user name here}"
		//si.ApplicationName = "{application's name}"
		xlWorkbook.SummaryInformation = si;
		CreateDateStyles();
	}

	/// <summary>
	/// See if the next row has data. Return False if end of sheet.
	/// </summary>
	/// <returns>True if read ok</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public bool ReadRow()
	{
        ReadRow(-1);
    }
    
    /// <summary>
	/// See if the next row has data. Return False if end of sheet.
	/// </summary>
	/// <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>True if read ok</returns>
	/// <remarks></remarks>
	public bool ReadRow(int Row)
	{
		try {
			if (Row > -1) {
				CurrentRow = Row;
			} else {
				CurrentRow += 1;
			}
			xlRow = xlSheet.GetRow(CurrentRow);
			if ((xlRow == null)) {
				return false;
			} else {
				return true;
			}
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			return false;
		}
	}

	/// <summary>
	/// Tests if we can read the next cell (it exists)
	/// </summary>
	/// <returns>True if the cell exists</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public bool CanReadCell()
	{
      CanReadCell(-1, -1);
   }	
	
	/// <summary>
	/// Tests if we can read the next cell (it exists)
	/// </summary>
	/// <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>True if the cell exists</returns>
	/// <remarks></remarks>
	public bool CanReadCell(int Column, int Row)
	{
		if (Row > -1)
			xlRow = xlSheet.GetRow(Row);
		if ((xlRow == null))
			return false;
		if (Column > -1) {
			if (Column > xlRow.PhysicalNumberOfCells - 1)
				return false;
		} else {
			if (CurrentColumn + 1 > xlRow.PhysicalNumberOfCells - 1)
				return false;
		}
		return true;
	}

	/// <summary>
	/// Read the value at the next cell (or at the specified cell). 
	/// </summary>
	/// <returns>Value or Nothing/Empty if read error</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public object ReadCell()
	{
      ReadCell(-1, -1);
	}
	
	/// <summary>
	/// Read the value at the next cell (or at the specified cell). 
	/// </summary>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>Value or Nothing/Empty if read error</returns>
	/// <remarks></remarks>
	public object ReadCell(int Column, int Row)
	{
		if (Row > -1) {
			CurrentRow = Row;
			xlRow = xlSheet.GetRow(Row);
		}
		if ((xlRow == null))
			return (ReturnEmptyOnError ? string.Empty : null);
		if (Column > -1) {
			CurrentColumn = Column;
		} else {
			CurrentColumn += 1;
		}
		if (CurrentColumn > xlRow.PhysicalNumberOfCells - 1) {
			return (ReturnEmptyOnError ? string.Empty : null);
		}

		if ((xlRow.GetCell(CurrentColumn) == null)) {
			return string.Empty;
		} else {
			switch (xlRow.GetCell(CurrentColumn).CellType) {
				case CellType.STRING:
					return xlRow.GetCell(CurrentColumn).StringCellValue;
				case CellType.NUMERIC:
					return xlRow.GetCell(CurrentColumn).NumericCellValue;
				case CellType.BOOLEAN:
					return xlRow.GetCell(CurrentColumn).BooleanCellValue;
				case CellType.FORMULA:
					//force to string
					return xlRow.GetCell(CurrentColumn).StringCellValue;
				default:
					return (ReturnEmptyOnError ? string.Empty : null);
			}
		}
	}

	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as Integer type. Strings are attempted to be converted to Integer
	/// </summary>
	/// <returns>Value or 0 if read error</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public int ReadCellAsInteger()
	{
 	    ReadCellAsInteger(-1, -1);
	}
	
	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as Integer type. Strings are attempted to be converted to Integer
	/// </summary>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>Value or 0 if read error</returns>
	/// <remarks></remarks>
	public int ReadCellAsInteger(int Column, int Row)
	{
		object @int = ReadCell(Column, Row);
		if (object.ReferenceEquals(@int.GetType(), typeof(int))) {
			return @int;
		} else if (object.ReferenceEquals(@int.GetType(), typeof(string))) {
			int xlInteger = 0;
			int.TryParse(@int, out xlInteger);
			return xlInteger;
		} else {
			return 0;
		}
	}

	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as Double type. Strings are attempted to be converted to Double
	/// </summary>
	/// <returns>Value or 0 if read error</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public double ReadCellAsDouble()
	{
 	    ReadCellAsDouble(-1, -1);
	}
	
	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as Double type. Strings are attempted to be converted to Double
	/// </summary>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>Value or 0 if read error</returns>
	/// <remarks></remarks>
	public double ReadCellAsDouble(int Column, int Row)
	{
		object dbl = ReadCell(Column, Row);
		if (object.ReferenceEquals(dbl.GetType(), typeof(double))) {
			return dbl;
		} else if (object.ReferenceEquals(dbl.GetType(), typeof(string))) {
			double xlDouble = 0;
			double.TryParse(dbl, out xlDouble);
			return xlDouble;
		} else {
			return 0;
		}
	}

	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as DateTime type. Strings are attempted to be converted to DateTime
	/// </summary>
	/// <returns>Value or 0 if read error</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public DateTime ReadCellAsDate()
	{
 	    ReadCellAsDate(-1, -1);
	}
	
	/// <summary>
	/// Read the value at the next cell (or at the specified cell) as DateTime type. Strings are attempted to be converted to DateTime
	/// </summary>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>Value or MinValue if read error</returns>
	/// <remarks></remarks>   
	public DateTime ReadCellAsDate(int Column, int Row)
	{
		object dbl = ReadCell(Column, Row);
		if (object.ReferenceEquals(dbl.GetType(), typeof(double))) {
			return xlRow.GetCell(CurrentColumn).DateCellValue;
		} else if (object.ReferenceEquals(dbl.GetType(), typeof(string))) {
			DateTime xlDate = default(DateTime);
			DateTime.TryParse(dbl, out xlDate);
			return xlDate;
		} else {
			return DateTime.MinValue;
		}
	}

	/// <summary>
	/// Write the value at the next cell (or at the specified cell). 
	/// </summary>
	/// <returns>True if ok</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public bool WriteCell(object Value)
	{
	   WriteCell(Value, -1, -1);
	}
	
	/// <summary>
	/// Write the value at the next cell (or at the specified cell). 
	/// </summary>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>True if ok</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public bool WriteCell(object Value, int Column, int Row)
	{
		try {
			if (Row > -1) {
				CurrentRow = Row;
				xlRow = xlSheet.GetRow(Row);
				if ((xlRow == null)) {
					xlRow = xlSheet.CreateRow(Row);
					CurrentColumn = -1;
				}
			} else {
				if ((xlRow == null)) {
					if (CurrentRow == -1)
						CurrentRow = 0;
					xlRow = xlSheet.CreateRow(CurrentRow);
					CurrentColumn = -1;
				}
			}
			if (Column > -1) {
				CurrentColumn = Column;
			} else {
				CurrentColumn += 1;
			}
			if (CurrentColumn > xlRow.PhysicalNumberOfCells - 1)
				xlCell = xlRow.CreateCell(CurrentColumn);
			if (object.ReferenceEquals(Value.GetType(), typeof(string))) {
				xlCell.SetCellType(CellType.STRING);
			} else if (object.ReferenceEquals(Value.GetType(), typeof(int)) | object.ReferenceEquals(typeof(ValueType), typeof(decimal)) | object.ReferenceEquals(typeof(ValueType), typeof(double))) {
				xlCell.SetCellType(CellType.NUMERIC);
			} else if (object.ReferenceEquals(Value.GetType(), typeof(DateTime))) {
				if (!Information.IsDBNull(Value)) {
					if (Value.TimeOfDay.TotalMilliseconds == 0) {
						xlCell.CellStyle = CellStyleDate;
					} else {
						xlCell.CellStyle = CellStyleDateTime;
					}
					xlCell.SetCellValue(Value);
				}
			} else if (object.ReferenceEquals(Value.GetType(), typeof(bool))) {
				xlCell.SetCellType(CellType.BOOLEAN);
			}
			xlCell.SetCellValue(Value);
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			return false;
		}
	}

	/// <summary>
	/// Write the date/teime value at the next cell (or at the specified cell) with the specified format. 
	/// </summary>
	/// <param name="Value">Date/Time value</param>
	/// <param name="DateTimeFormat">Format to display the date/time in. Nothing/Empty=default thread culture. ie: "dd/mm/yyyy hh:mm:ss AM/PM"</param>
	/// <returns>True if ok</returns>
	/// <remarks>Overload added for older c# vers</remarks>
	public bool WriteCellAsDate(DateTime Value, string DateTimeFormat)
	{
	   WriteCellAsDate(Value, DateTimeFormat, -1, -1);
	}
	
	/// <summary>
	/// Write the date/teime value at the next cell (or at the specified cell) with the specified format. 
	/// </summary>
	/// <param name="Value">Date/Time value</param>
	/// <param name="DateTimeFormat">Format to display the date/time in. Nothing/Empty=default thread culture. ie: "dd/mm/yyyy hh:mm:ss AM/PM"</param>
	/// <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
	/// <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
	/// <returns>True if ok</returns>
	/// <remarks></remarks>
	public bool WriteCellAsDate(DateTime Value, string DateTimeFormat, int Column, int Row)
	{
		try {
			if (Row > -1) {
				CurrentRow = Row;
				xlRow = xlSheet.GetRow(Row);
				if ((xlRow == null)) {
					xlRow = xlSheet.CreateRow(Row);
					CurrentColumn = -1;
				}
			} else {
				if ((xlRow == null)) {
					if (CurrentRow == -1)
						CurrentRow = 0;
					xlRow = xlSheet.CreateRow(CurrentRow);
					CurrentColumn = -1;
				}
			}
			if (Column > -1) {
				CurrentColumn = Column;
			} else {
				CurrentColumn += 1;
			}
			if (CurrentColumn > xlRow.PhysicalNumberOfCells - 1)
				xlCell = xlRow.CreateCell(CurrentColumn);
			if (!Information.IsDBNull(Value)) {
				if (string.IsNullOrEmpty(DateTimeFormat)) {
					if (Value.TimeOfDay.TotalMilliseconds == 0) {
						xlCell.CellStyle = CellStyleDate;
					} else {
						xlCell.CellStyle = CellStyleDateTime;
					}
				} else {
					if (!CellStyleList.ContainsKey(DateTimeFormat)) {
						HSSFDataFormat format = xlWorkbook.CreateDataFormat();
						HSSFCellStyle xlStyle = xlWorkbook.CreateCellStyle();
						xlStyle.DataFormat = format.GetFormat(DateTimeFormat);
						CellStyleList.Add(DateTimeFormat, xlStyle);
					}
					xlCell.CellStyle = CellStyleList(DateTimeFormat);
				}
				xlCell.SetCellValue(Value);
			}
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			return false;
		}
	}

	/// <summary>
	/// Set the cell style of the last cell
	/// </summary>
	/// <param name="Bold">Bold font. Default = False</param>
	/// <param name="Align">Cell alignment. Default = NPOI.SS.UserModel.HorizontalAlignment.GENERAL</param>
	/// <param name="Width">Cell Width. Default = 10</param>
	/// <param name="CellSpan">Cell span. Default = 1</param>
	/// <param name="FGColour">Font colour. none=System.Drawing.Color.Empty</param>
	/// <param name="BGColour">Background colour. none=System.Drawing.Color.Empty</param>
	/// <param name="Border">Thin black border. Default = False</param>
	/// <returns>True if ok</returns>
	/// <remarks></remarks>
	public bool FormatLastCell(bool Bold, HorizontalAlignment Align, int Width, int CellSpan, System.Drawing.Color FGColour, System.Drawing.Color BGColour, bool Border)
	{
		string strStyleIndex = string.Concat(Bold.ToString(), Align.ToString, FGColour.ToString, BGColour.ToString, Border.ToString());
		//reuse existing styles (avoid limit to number of new styles)
		if (!CellStyleList.ContainsKey(strStyleIndex)) {
			HSSFCellStyle xlStyle = xlCell.CellStyle;
			if ((xlStyle == null))
				xlStyle = xlWorkbook.CreateCellStyle();
			if (Bold | FGColour != System.Drawing.Color.Empty) {
				HSSFFont xlFont = xlWorkbook.CreateFont();
				if (FGColour != System.Drawing.Color.Empty)
					xlFont.Color = GetXLColourFG(FGColour);
				if (Bold)
					xlFont.Boldweight = FontBoldWeight.BOLD;
				xlStyle.SetFont(xlFont);
			}
			if (BGColour != System.Drawing.Color.Empty) {
				if (BGColour != System.Drawing.Color.Empty)
					xlStyle.FillForegroundColor = GetXLColourBG(BGColour);
				xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND;
			}
			if (Border) {
				xlStyle.BorderTop = CellBorderType.THIN;
				xlStyle.BorderLeft = CellBorderType.THIN;
				xlStyle.BorderRight = CellBorderType.THIN;
				xlStyle.BorderBottom = CellBorderType.THIN;
			}
			if (Align != HorizontalAlignment.GENERAL) {
				xlStyle.Alignment = Align;
			}
			CellStyleList.Add(strStyleIndex, xlStyle);
		}
		xlCell.CellStyle = CellStyleList(strStyleIndex);
		if (CellSpan > 1) {
			NPOI.SS.Util.CellRangeAddress LastMergedCells = new NPOI.SS.Util.CellRangeAddress(CurrentRow, CurrentRow, CurrentColumn, CurrentColumn + CellSpan);
			xlSheet.AddMergedRegion(LastMergedCells);
			if (Border) {
				//xlSheet.SetEnclosedBorderOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
				xlSheet.SetBorderTopOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
				xlSheet.SetBorderBottomOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
				xlSheet.SetBorderLeftOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
				xlSheet.SetBorderRightOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
			}
		}
		xlSheet.SetColumnWidth(CurrentColumn, Width * 256);
	}

	/// <summary>
	/// Creates a new row for writing. Use this if using WriteCell with no row/column specified.
	/// </summary>
	/// <returns>True if ok</returns>
	/// <remarks></remarks>
	public bool WriteNewRow()
	{
		CurrentRow += 1;
		try {
			xlRow = xlSheet.CreateRow(CurrentRow);
			CurrentColumn = -1;
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			return false;
		}
	}

	/// <summary>
	/// Save the workbook to a new file
	/// </summary>
	/// <param name="XlsFilePath">Absolute path and filename for the .xls</param>
	/// <returns>True if saved</returns>
	/// <remarks></remarks>
	public bool Save(string XlsFilePath)
	{
		try {
			if (File.Exists(XlsFilePath))
				File.Delete(XlsFilePath);
			using (FileStream fs = new FileStream(XlsFilePath, FileMode.Create)) {
				xlWorkbook.Write(fs);
				fs.Flush();
				fs.Close();
			}
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
			return false;
		}
	}

	/// <summary>
	/// Save the workbook to a stream or filestream
	/// </summary>
	/// <param name="XlsFileStream">Stream or FileStream to save the .xls to</param>
	/// <returns>True if saved</returns>
	/// <remarks></remarks>
	public bool Save(ref Stream XlsFileStream)
	{
		try {
			xlWorkbook.Write(XlsFileStream);
			return true;
		} catch (Exception ex) {
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
			return false;
		}
	}


	/// <summary>
	/// Save a GridView Control to an Excel 97+ BIFF .xls file
	/// </summary>
	/// <param name="GridViewControl">Databound (DataBind already called) GridView Control</param>
	/// <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
	/// <returns>True if successful</returns>
	/// <remarks></remarks>
	public bool GridViewToXLS(ref GridView GridViewControl, string FileToSave)
	{
		try {
			if (IO.File.Exists(FileToSave))
				IO.File.Delete(FileToSave);
		} catch {
			//HttpContext.Current.Trace.Warn("Can not delete previous file")
		}
		if (!FileToSave.Contains("."))
			FileToSave += ".xls";
		FileInfo fiSave = new FileInfo(FileToSave);
		string ReportName = fiSave.Name.Substring(0, fiSave.Name.IndexOf("."));
		ReportName = Regex.Replace(ReportName, "[^A-Z]", string.Empty, RegexOptions.IgnoreCase);
		bool blnNoError = true;

		xlWorkbook = new HSSFWorkbook();
		CreateStandardProperties(ReportName);

		xlSheet.DisplayGridlines = false;
		HSSFCellStyle xlStyle = default(HSSFCellStyle);
		HSSFFont xlFont = default(HSSFFont);
		Int16[] aryWidths = null;
		aryWidths = new short[1];
		CurrentRow = 0;
		//xlWorkbook.NumCellStyles
		//xlWorkbook.GetCellStyleAt()
		NPOI.SS.Util.CellRangeAddress LastMergedCells = default(NPOI.SS.Util.CellRangeAddress);
		bool blnIsDate = false;
		bool blnIsDecimal = false;

		//Write tables
		Regex regExHTML = new Regex("<([^>]*)>[^<]*<\\/\\1>|<[^>]+/>|<[^>]+>", RegexOptions.Multiline | RegexOptions.IgnoreCase);
		System.Web.UI.WebControls.TableCell TableCell = default(System.Web.UI.WebControls.TableCell);

		//Header
		if ((GridViewControl.HeaderRow != null)) {
			xlRow = xlSheet.CreateRow(CurrentRow);
			xlStyle = xlWorkbook.CreateCellStyle();
			xlFont = xlWorkbook.CreateFont();
			xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index;
			xlFont.Boldweight = FontBoldWeight.BOLD;
			xlStyle.SetFont(xlFont);
			xlStyle.FillForegroundColor = GetXLColourBG(System.Drawing.ColorTranslator.FromHtml("#4666ad"));
			xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND;
			aryWidths = new short[GridViewControl.HeaderRow.Cells.Count];
			foreach ( TableCell in GridViewControl.HeaderRow.Cells) {
				if (GridViewControl.Columns(CurrentColumn).Visible) {
					xlCell = xlRow.CreateCell(CurrentColumn);
					xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(Regex.Replace(TableCell.Text, "<[^>]*>", " ")));
					xlCell.CellStyle = xlStyle;
					if (TableCell.Text.Length + 1 > aryWidths[CurrentColumn]) {
						aryWidths[CurrentColumn] = TableCell.Text.Length + 1;
					}
					CurrentColumn += 1;
				}
			}
			CurrentRow += 1;
		}

		//Data rows
		int intCol = 0;
		int intRow = 0;
		int intValue = 0;
		int intCellSize = 0;
		int ActualCol = 0;
		decimal decValue = default(decimal);
		string strValue = null;
		GridViewRow GridViewRowControl = default(GridViewRow);
		for (intRow = 0; intRow <= GridViewControl.Rows.Count - 1; intRow++) {
			GridViewRowControl = GridViewControl.Rows(intRow);
			xlRow = xlSheet.CreateRow(CurrentRow);
			if (aryWidths.GetUpperBound(0) < GridViewRowControl.Cells.Count)
				Array.Resize(ref aryWidths, GridViewRowControl.Cells.Count);
			ActualCol = 0;
			for (intCol = 0; intCol <= GridViewRowControl.Cells.Count - 1; intCol++) {
				TableCell = GridViewRowControl.Cells(intCol);
				if (GridViewControl.Columns(intCol).Visible) {
					if (TableCell.Controls.Count > 0) {
						strValue = string.Empty;
						foreach (Control cntl in TableCell.Controls) {
							if (object.ReferenceEquals(cntl.GetType, typeof(DataBoundLiteralControl))) {
								strValue += ((DataBoundLiteralControl)cntl).Text.Trim;
							} else if (object.ReferenceEquals(cntl.GetType, typeof(CheckBox))) {
								strValue += ((CheckBox)cntl).Checked.ToString;
							} else if (object.ReferenceEquals(cntl.GetType, typeof(SymTextBox))) {
								strValue += ((SymTextBox)cntl).Text;
							} else if (object.ReferenceEquals(cntl.GetType, typeof(DropDownList))) {
								strValue += ((DropDownList)cntl).SelectedItem.Text;
							} else if (object.ReferenceEquals(cntl.GetType, typeof(ListBox))) {
								foreach (ListItem li in ((ListBox)cntl).Items) {
									if (li.Selected) {
										if (strValue != string.Empty)
											strValue += ", ";
										strValue += li.Text;
									}
								}
							} else {
								strValue += TableCell.Text;
							}
						}
					} else {
						strValue = TableCell.Text;
					}

					//Merge/span cells
					if (TableCell.ColumnSpan > 1 | TableCell.RowSpan > 1) {
						LastMergedCells = new NPOI.SS.Util.CellRangeAddress(CurrentRow, CurrentRow, ActualCol, ActualCol + TableCell.ColumnSpan - 1);
						xlSheet.AddMergedRegion(LastMergedCells);
					} else {
						LastMergedCells = null;
					}
					//xlCell = xlRow.Cells.Add()
					xlCell = xlRow.CreateCell(ActualCol);

					//reuse existing styles (avoid limit to number of new styles)
					blnIsDate = Information.IsDate(strValue);
					blnIsDecimal = strValue.Contains(".") && decimal.TryParse(strValue, System.Globalization.NumberStyles.Currency, null, out decValue);
					if (blnIsDate | blnIsDecimal | TableCell.Font.Bold | TableCell.ForeColor != Color.Empty | TableCell.BackColor != Color.Empty | GridViewControl.GridLines > GridLines.None | TableCell.BorderWidth.Value > 0 | TableCell.HorizontalAlign != HorizontalAlign.NotSet) {
						string strStyle = string.Concat(blnIsDate.ToString(), (blnIsDate ? TableCell.Text.Contains(":").ToString : string.Empty), blnIsDecimal.ToString(), TableCell.Font.Bold.ToString, TableCell.ForeColor, TableCell.BackColor, GridViewControl.GridLines.ToString, TableCell.BorderWidth.Value.ToString, TableCell.HorizontalAlign.ToString);

						if ((LastMergedCells != null) && (GridViewControl.GridLines > GridLines.None | TableCell.BorderWidth.Value > 0)) {
							//xlSheet.SetEnclosedBorderOfRegion(LastMergedCells, CellBorderType.DASHED, NPOI.HSSF.Util.HSSFColor.BLACK.index)
							xlSheet.SetBorderTopOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
							xlSheet.SetBorderBottomOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
							xlSheet.SetBorderLeftOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
							xlSheet.SetBorderRightOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index);
						}

						if (!CellStyleList.ContainsKey(strStyle)) {
							xlStyle = xlWorkbook.CreateCellStyle();
							//Colour
							if (TableCell.Font.Bold | TableCell.ForeColor != Color.Empty) {
								xlFont = xlWorkbook.CreateFont();
								if (TableCell.ForeColor != Color.Empty)
									xlFont.Color = GetXLColourFG(TableCell.ForeColor);
								if (TableCell.Font.Bold)
									xlFont.Boldweight = FontBoldWeight.BOLD;
								xlStyle.SetFont(xlFont);
							}
							if (TableCell.BackColor != Color.Empty) {
								xlStyle.FillForegroundColor = GetXLColourBG(TableCell.BackColor);
								xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND;
							}

							if (GridViewControl.GridLines > GridLines.None | TableCell.BorderWidth.Value > 0) {
								//Border/Grid
								xlStyle.BorderTop = CellBorderType.THIN;
								xlStyle.BorderLeft = CellBorderType.THIN;
								xlStyle.BorderRight = CellBorderType.THIN;
								xlStyle.BorderBottom = CellBorderType.THIN;
							}

							//Alignment
							if (TableCell.HorizontalAlign != HorizontalAlign.NotSet) {
								switch (TableCell.HorizontalAlign) {
									case HorizontalAlign.Left:
										xlStyle.Alignment = HorizontalAlignment.LEFT;
										break;
									case HorizontalAlign.Right:
										xlStyle.Alignment = HorizontalAlignment.RIGHT;
										break;
									case HorizontalAlign.Center:
										xlStyle.Alignment = HorizontalAlignment.CENTER;
										break;
								}
							}

							if (blnIsDate) {
								//Date format
								if (strValue.Contains(":")) {
									xlStyle.DataFormat = FormatDateTime;
									intCellSize = DateTimeFormat.Length;
								} else {
									xlStyle.DataFormat = FormatDate;
									intCellSize = DateFormat.Length;
								}
							} else if (blnIsDecimal) {
								//Common decimal styles
								if (TableCell.Text.Length - TableCell.Text.IndexOf(".") == 3) {
									xlStyle.DataFormat = HSSFDataFormat.GetBuiltinFormat("0.00");
								} else if (TableCell.Text.Length - TableCell.Text.IndexOf(".") == 2) {
									xlStyle.DataFormat = HSSFDataFormat.GetBuiltinFormat("0.0");
								}
							}

							CellStyleList.Add(strStyle, xlStyle);
						}
						xlCell.CellStyle = CellStyleList(strStyle);
					}

					if (int.TryParse(strValue, System.Globalization.NumberStyles.Integer | System.Globalization.NumberStyles.AllowThousands, null, out intValue)) {
						xlCell.SetCellType(CellType.NUMERIC);
						xlCell.SetCellValue(intValue);
						intCellSize = strValue.Length;
					} else if (blnIsDecimal) {
						xlCell.SetCellType(CellType.NUMERIC);
						xlCell.SetCellValue(decValue);
						intCellSize = strValue.Length;
					} else if (blnIsDate) {
						xlCell.SetCellValue(strValue);
					} else if (regExHTML.IsMatch(strValue)) {
						strValue = System.Web.HttpUtility.HtmlDecode(Regex.Replace(strValue, "<[^>]*>", string.Empty));
						if (decimal.TryParse(strValue, out decValue)) {
							xlCell.SetCellType(CellType.NUMERIC);
							xlCell.SetCellValue(decValue);
						} else {
							xlCell.SetCellValue(strValue);
						}
						intCellSize = strValue.Length;
					} else {
						xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(strValue));
						intCellSize = xlCell.StringCellValue.Length;
					}

					if (intCellSize > aryWidths[intCol] && GridViewRowControl.Cells(intCol).ColumnSpan < 2) {
						aryWidths[intCol] = intCellSize;
					}
					ActualCol += (TableCell.ColumnSpan == 0 ? 1 : TableCell.ColumnSpan);
				}
			}
			CurrentRow += 1;
		}

		//Footer
		if ((GridViewControl.FooterRow != null)) {
			xlRow = xlSheet.CreateRow(CurrentRow);
			xlStyle = xlWorkbook.CreateCellStyle();
			xlFont = xlWorkbook.CreateFont();
			xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index;
			xlFont.Boldweight = FontBoldWeight.BOLD;
			xlStyle.SetFont(xlFont);
			xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.GREY_50_PERCENT.index;
			xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND;
			CurrentColumn = 0;
			foreach ( TableCell in GridViewControl.FooterRow.Cells) {
				if (GridViewControl.Columns(CurrentColumn).Visible) {
					xlCell = xlRow.CreateCell(CurrentColumn);
					xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(Regex.Replace(TableCell.Text, "<[^>]*>", string.Empty)));
					xlStyle.Alignment = HorizontalAlignment.RIGHT;
					xlCell.CellStyle = xlStyle;
					if (TableCell.Text.Length + 1 > aryWidths[CurrentColumn]) {
						aryWidths[CurrentColumn] = TableCell.Text.Length + 1;
					}
					CurrentColumn += 1;
				}
			}
		}

		//Set column widths (NPOI auto-width can be ugly)
		for (int intColumn = 0; intColumn <= aryWidths.GetUpperBound(0); intColumn++) {
			if (aryWidths[intColumn] > 50) {
				xlSheet.SetColumnWidth(intColumn, 50 * 256);
			} else if (aryWidths[intColumn] < 10) {
				xlSheet.SetColumnWidth(intColumn, 10 * 256);
			} else {
				xlSheet.SetColumnWidth(intColumn, (aryWidths[intColumn] + 1) * 256);
			}
		}

		try {
			if (FileToSave.IndexOf(":") < 1) {
				//just filename to stream to browser and force it to download the file (and launch Excel or save, not display in the browser)
				MemoryStream ms = new MemoryStream();
				xlWorkbook.Write(ms);
				HttpContext.Current.Response.Clear();
				HttpContext.Current.Response.AddHeader("Content-Disposition", "attachment; filename=" + FileToSave);
				HttpContext.Current.Response.AddHeader("Content-Length", ms.Length.ToString());
				HttpContext.Current.Response.ContentType = "application/octet-stream";
				HttpContext.Current.Response.BinaryWrite(ms.GetBuffer());
				HttpContext.Current.Response.Flush();
				HttpContext.Current.Response.End();
			} else {
				using (FileStream fs = new FileStream(FileToSave, FileMode.Create)) {
					xlWorkbook.Write(fs);
					fs.Close();
				}
			}
		} catch (Exception ex) {
			blnNoError = false;
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
		}
		return blnNoError;
	}

	/// <summary>
	/// Save a Table to an Excel 97+ BIFF .xls file
	/// </summary>
	/// <param name="SourceData">Source table</param>
	/// <param name="SourceColOrder">Source column name order. Nothing if auto order (per table). Required if SourceColText specified</param>
	/// <param name="SourceColText">Source column header text. Nothing if auto names (per table). Required if SourceColOrder specified</param>
	/// <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
	/// <param name="IncludeHeader">Write column names</param>
	/// <returns>True if successful</returns>
	/// <remarks></remarks>
	public bool TableToXLS(ref DataTable SourceData, ref string[] SourceColOrder, ref string[] SourceColText, bool IncludeHeader, string FileToSave)
	{
		try {
			if (IO.File.Exists(FileToSave))
				IO.File.Delete(FileToSave);
		} catch {
			//HttpContext.Current.Trace.Warn("Can not delete previous file")
		}
		FileInfo fiSave = new FileInfo(FileToSave);
		string SheetName = fiSave.Name.Substring(0, fiSave.Name.IndexOf("."));
		SheetName = Regex.Replace(SheetName, "[^A-Z]", string.Empty, RegexOptions.IgnoreCase);
		bool blnNoError = true;

		xlWorkbook = new HSSFWorkbook();
		CreateStandardProperties(SheetName);

		xlSheet = xlWorkbook.CreateSheet(SheetName);
		xlSheet.DisplayGridlines = true;
		HSSFCellStyle xlStyle = default(HSSFCellStyle);
		Int16[] aryWidths = null;
		aryWidths = new short[SourceData.Columns.Count];
		CurrentRow = 0;

		//Write header
		if (IncludeHeader) {
			xlRow = xlSheet.CreateRow(CurrentRow);
			xlStyle = xlWorkbook.CreateCellStyle();
			HSSFFont xlFont = xlWorkbook.CreateFont();
			xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index;

			xlFont.Boldweight = FontBoldWeight.BOLD;
			xlStyle.SetFont(xlFont);
			xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.ROYAL_BLUE.index;
			xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND;

			if ((SourceColText == null)) {
				//Auto names
				SourceColOrder = new string[SourceData.Columns.Count];
				foreach (DataColumn dcCol in SourceData.Columns) {
					SourceColOrder[CurrentColumn] = dcCol.ColumnName;
					xlCell = xlRow.CreateCell(CurrentColumn);
					xlCell.SetCellValue(dcCol.ColumnName.Replace("_", " "));
					xlCell.CellStyle = xlStyle;
					if (dcCol.ColumnName.Length + 1 > aryWidths[CurrentColumn]) {
						aryWidths[CurrentColumn] = dcCol.ColumnName.Length + 1;
					}
					CurrentColumn += 1;
				}
			} else {
				//Specified names
				foreach (string strName in SourceColText) {
					xlCell = xlRow.CreateCell(CurrentColumn);
					xlCell.SetCellValue(strName);
					xlCell.CellStyle = xlStyle;
					if (strName.Length + 1 > aryWidths[CurrentColumn]) {
						aryWidths[CurrentColumn] = strName.Length + 1;
					}
					CurrentColumn += 1;
				}
			}
		}

		if ((SourceColOrder == null)) {
			//Get column names if not specified
			CurrentColumn = 0;
			SourceColOrder = new string[SourceData.Columns.Count];
			foreach (DataColumn dcCol in SourceData.Columns) {
				SourceColOrder[CurrentColumn] = dcCol.ColumnName;
				CurrentColumn += 1;
			}
		}

		//Write data rows
		object SourceDataCell = null;
		foreach (DataRow drvRow in SourceData.Rows) {
			CurrentRow += 1;
			CurrentColumn = 0;
			xlRow = xlSheet.CreateRow(CurrentRow);
			//For CurrentColumn = 0 To SourceData.Columns.Count - 1
			foreach (string strColName in SourceColOrder) {
				if (strColName != string.Empty) {
					SourceDataCell = drvRow(strColName);
					xlCell = xlRow.CreateCell(CurrentColumn);
					if (object.ReferenceEquals(SourceDataCell.GetType(), typeof(System.Int32)) | object.ReferenceEquals(SourceDataCell.GetType(), typeof(System.Int64))) {
						xlCell.SetCellType(CellType.NUMERIC);
						if (!Information.IsDBNull(SourceDataCell))
							xlCell.SetCellValue(SourceDataCell);
					} else if (object.ReferenceEquals(SourceDataCell.GetType(), typeof(System.DateTime))) {
						if (!Information.IsDBNull(SourceDataCell)) {
							if (SourceDataCell.TimeOfDay.TotalMilliseconds == 0) {
								xlCell.CellStyle = CellStyleDate;
							} else {
								xlCell.CellStyle = CellStyleDateTime;
							}
							xlCell.SetCellValue(SourceDataCell);
						}
					} else if (object.ReferenceEquals(SourceDataCell.GetType(), typeof(System.Double)) | object.ReferenceEquals(SourceDataCell.GetType(), typeof(System.Decimal))) {
						xlCell.SetCellType(CellType.NUMERIC);
						if (!Information.IsDBNull(SourceDataCell))
							xlCell.SetCellValue(SourceDataCell);
					} else {
						xlCell.SetCellValue(SourceDataCell.ToString());
					}
					if (SourceDataCell.ToString().Length > aryWidths[CurrentColumn]) {
						aryWidths[CurrentColumn] = SourceDataCell.ToString().Length;
					}
					CurrentColumn += 1;
				}
			}
		}

		//Set column widths (NPOI auto-width can be ugly)
		for (int intColumn = 0; intColumn <= aryWidths.GetUpperBound(0); intColumn++) {
			if (aryWidths[intColumn] > 50) {
				xlSheet.SetColumnWidth(intColumn, 50 * 256);
			} else if (aryWidths[intColumn] < 10) {
				xlSheet.SetColumnWidth(intColumn, 10 * 256);
			} else {
				xlSheet.SetColumnWidth(intColumn, (aryWidths[intColumn] + 1) * 256);
			}
		}

		try {
			if (FileToSave.IndexOf(":") < 1) {
				//just filename to stream to browser and force it to download the file (and launch Excel or save, not display in the browser)
				MemoryStream ms = new MemoryStream();
				xlWorkbook.Write(ms);
				HttpContext.Current.Response.Clear();
				HttpContext.Current.Response.AddHeader("Content-Disposition", "attachment; filename=" + FileToSave);
				HttpContext.Current.Response.AddHeader("Content-Length", ms.Length.ToString());
				HttpContext.Current.Response.ContentType = "application/octet-stream";
				HttpContext.Current.Response.BinaryWrite(ms.GetBuffer());
				HttpContext.Current.Response.Flush();
				HttpContext.Current.Response.End();
			} else {
				using (FileStream fs = new FileStream(FileToSave, FileMode.Create)) {
					xlWorkbook.Write(fs);
					fs.Close();
				}
			}
		} catch (Exception ex) {
			blnNoError = false;
			ErrorMessage = ex.Message;
			//HttpContext.Current.Trace.Warn(ex.ToString)
		}
		return blnNoError;
	}

	private void CreateDateStyles()
	{
		//Create a default date and date/time format and style for general use
		HSSFDataFormat format = xlWorkbook.CreateDataFormat();
		DateFormat = Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortDatePattern.ToLower;
		FormatDate = format.GetFormat(DateFormat);
		format = xlWorkbook.CreateDataFormat();
		CellStyleDate = xlWorkbook.CreateCellStyle();
		CellStyleDate.DataFormat = FormatDate;
		DateTimeFormat = string.Concat(DateFormat, " ", Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortTimePattern.ToLower.Replace("tt", "AM/PM"));
		FormatDateTime = format.GetFormat(DateTimeFormat);
		CellStyleDateTime = xlWorkbook.CreateCellStyle();
		CellStyleDateTime.DataFormat = FormatDateTime;
	}

	private short GetXLColourFG(System.Drawing.Color SystemColour)
	{
		//Lookup RGB from .NET system colour in Excel pallete - or nearest match. Font colour in 1.2.3 has a bug and does not support custom entries
		if (SystemColour == System.Drawing.Color.Empty)
			return NPOI.HSSF.Util.HSSFColor.COLOR_NORMAL;
		//COLOR_NORMAL=unspecified. 'alt= AUTOMATIC.index 
		HSSFPalette XlPalette = xlWorkbook.GetCustomPalette();
		NPOI.HSSF.Util.HSSFColor XlColour = XlPalette.FindColor(SystemColour.R, SystemColour.G, SystemColour.B);
		if ((XlColour == null)) {
			XlColour = XlPalette.FindSimilarColor(SystemColour.R, SystemColour.G, SystemColour.B);
			return XlColour.GetIndex();
		} else {
			return XlColour.GetIndex();
		}
	}

	private short GetXLColourBG(System.Drawing.Color SystemColour)
	{
		//Lookup RGB from .NET system colour in Excel pallete - or create a new entry (get nearest if palette full). Return the XL palette index.
		if (SystemColour == System.Drawing.Color.Empty)
			return NPOI.HSSF.Util.HSSFColor.COLOR_NORMAL;
		//COLOR_NORMAL=unspecified. 'alt= AUTOMATIC.index 
		HSSFPalette XlPalette = xlWorkbook.GetCustomPalette();
		NPOI.HSSF.Util.HSSFColor XlColour = XlPalette.FindColor(SystemColour.R, SystemColour.G, SystemColour.B);
		if ((XlColour == null)) {
			//Available colour palette entries: 65 to 32766 (0-64=standard palette; 64=auto, 32767=unspecified)
			if (NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE < 255) {
				if (NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE < 67) {
					//Move start pointer up. First Entry seems to fail, so add dummy.
					NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE = 67;
					XlColour = XlPalette.AddColor(100, 100, 100);
					XlColour = XlPalette.AddColor(100, 100, 100);
				}
				NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE += 1;
				XlColour = XlPalette.AddColor(SystemColour.R, SystemColour.G, SystemColour.B);
			} else {
				XlColour = XlPalette.FindSimilarColor(SystemColour.R, SystemColour.G, SystemColour.B);
			}
			return XlColour.GetIndex();
		} else {
			return XlColour.GetIndex();
		}
	}

}

//End Namespace
