'Excel XLS BIFF File Reader/Writer - Shell for NPOI 1.2.4
'v1.1 +TableToXLS, fix ReadCell exception if value is nothing
'v1.2 Fixed IsNothing(xlRow) in ReadCell/CanReadCell
'v1.3 Fixed visible & no sort in TableToXLS
'v1.4 Update to NPOI1.2.2 (new Enums). Colour
'v1.5 Updated to NPOI1.2.3 (NPOI.SS.UserModel). Date/Time data format set properly
'v1.6 Fixed FontBoldWeight
'v1.7 Much better style management
'v2.0 Open and save to streams. Swapped Col and Row parameters. Fixed formula type
'v2.1 GetXLColourBG supports adding custom colours to Excel. GetXLColourFG for font colour - gets nearest colour due to palette bug in 1.2.3. Fixed merged cell border
'v2.2 Fix GetXLColourBG in 1.2.4, Fix setting of date and decimal styles in TableControlsToXLS, GridViewToXLS

Imports System
Imports System.Web
Imports System.Data
Imports System.Collections
Imports Microsoft.VisualBasic
Imports System.IO
Imports NPOI.HSSF.UserModel
Imports NPOI.SS.UserModel
Imports NPOI.HPSF
Imports NPOI.POIFS.FileSystem

'Namespace Bean.Excel

    'Sample Code (to write a new file a read it back again):
    'Dim xr As New NPOIExcel
    'xr.Open()
    'xr.WriteNewRow()
    'xr.WriteCell("text")
    'xr.WriteCell(1234)
    'xr.WriteCell(Now)
    'xr.WriteCellAsDate(Now, "dd/mm/yy")
    'xr.FormatLastCell(True, NPOI.SS.UserModel.HorizontalAlignment.RIGHT, 15, 2, NPOI.HSSF.Util.HSSFColor.BLUE.index, NPOI.HSSF.Util.HSSFColor.OLIVE_GREEN.index, True)
    'xr.Save("c:\web\sample data.xls")
    ''Read it back again
    'xr.Open("c:\web\sample data.xls")
    'While xr.ReadRow
    '    While xr.CanReadCell()
    '        lblMessage.Text &= "," & xr.ReadCell().ToString
    '    End While
    '    lblMessage.Text &= "<p/>"
    'End While

    ''' <summary>
    ''' Provides easy to use API for basic Excel 97+ BIFF file reading and writing. Exposes the NPOI Excel objects for advanced usage.
    ''' </summary>
    ''' <remarks></remarks>
    Public Class NPOIExcel
        Public ErrorMessage As String
        ''' <summary>
        ''' Return String.Empty or Nothing if error (such as no more rows/columns)
        ''' </summary>
        Public ReturnEmptyOnError As Boolean
        Public xlWorkbook As HSSFWorkbook
        Public xlSheet As HSSFSheet
        Public xlRow As HSSFRow
        Public xlCell As HSSFCell
        Public CurrentRow, CurrentColumn As Integer
        Public CellStyleDate, CellStyleDateTime As HSSFCellStyle
        Private CellStyleList As New Generic.Dictionary(Of String, HSSFCellStyle) 'for user defined styles. Reuse existing styles (avoid limit to number of new styles)
        Private FormatDate, FormatDateTime As Short
        Private DateFormat, DateTimeFormat As String

        ''' <summary>
        ''' Open the XLS file for reading OR create a new xls workbook. (Does not use file or stream until Save is called)
        ''' </summary>
        ''' <param name="XlsFilePath">Path to the file to open. Empty/not specified for new file</param>
        ''' <returns>True if opened successfully</returns>
        ''' <remarks></remarks>
        Public Function Open(Optional ByVal XlsFilePath As String = "") As Boolean
            Try
                If XlsFilePath = String.Empty Then
                    xlWorkbook = New HSSFWorkbook()
                    CreateStandardProperties("Sheet1")
                Else
                    If File.Exists(XlsFilePath) Then
                        Using fs As New FileStream(XlsFilePath, FileMode.Open)
                            xlWorkbook = New HSSFWorkbook(fs)
                            If xlWorkbook.NumberOfSheets > 0 Then
                                xlSheet = xlWorkbook.GetSheetAt(0)
                            Else
                                CreateStandardProperties(XlsFilePath)
                            End If
                            fs.Close()
                        End Using
                    Else
                        xlWorkbook = New HSSFWorkbook()
                        CreateStandardProperties(XlsFilePath)
                    End If
                End If
                CurrentRow = -1
                CurrentColumn = -1
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Open the XLS file stream for reading
        ''' </summary>
        ''' <param name="XlsFileStream">Stream or FileStream to open</param>
        ''' <returns>True if opened successfully</returns>
        ''' <remarks></remarks>
        Public Function Open(ByVal XlsFileStream As Stream) As Boolean
            Try
                xlWorkbook = New HSSFWorkbook(XlsFileStream)
                If XlsFileStream.Length = 0 Then
                    CreateStandardProperties("Sheet1")
                Else
                    xlSheet = xlWorkbook.GetSheetAt(0)
                End If
                CurrentRow = -1
                CurrentColumn = -1
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function

		Private Sub CreateStandardProperties(ByVal SheetName As String)
			'Create some standard properties and styles for new spreasheets
            xlSheet = xlWorkbook.CreateSheet(SheetName)
            Dim dsi As DocumentSummaryInformation = PropertySetFactory.CreateDocumentSummaryInformation()
            'dsi.Company = "{Company Name}"
            xlWorkbook.DocumentSummaryInformation = dsi
            Dim si As SummaryInformation = PropertySetFactory.CreateSummaryInformation()
            si.Title = SheetName
            'si.Author = "{user name here}"
            'si.ApplicationName = "{application's name}"
            xlWorkbook.SummaryInformation = si
            CreateDateStyles()
        End Sub

        ''' <summary>
        ''' See if the next row has data. Return False if end of sheet.
        ''' </summary>
        ''' <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if read ok</returns>
        ''' <remarks></remarks>
        Public Function ReadRow(Optional ByVal Row As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                Else
                    CurrentRow += 1
                End If
                xlRow = xlSheet.GetRow(CurrentRow)
                If IsNothing(xlRow) Then
                    Return False
                Else
                    Return True
                End If
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Tests if we can read the next cell (it exists)
        ''' </summary>
        ''' <param name="Row">If provided, skip to the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if the cell exists</returns>
        ''' <remarks></remarks>
        Public Function CanReadCell(Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Boolean
            If Row > -1 Then xlRow = xlSheet.GetRow(Row)
            If IsNothing(xlRow) Then Return False
            If Column > -1 Then
                If Column > xlRow.PhysicalNumberOfCells - 1 Then Return False
            Else
                If CurrentColumn + 1 > xlRow.PhysicalNumberOfCells - 1 Then Return False
            End If
            Return True
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell). 
        ''' </summary>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or Nothing/Empty if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCell(Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Object
            If Row > -1 Then
                CurrentRow = Row
                xlRow = xlSheet.GetRow(Row)
            End If
            If IsNothing(xlRow) Then Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
            If Column > -1 Then
                CurrentColumn = Column
            Else
                CurrentColumn += 1
            End If
            If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then
                Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
            End If

            If IsNothing(xlRow.GetCell(CurrentColumn)) Then
                Return String.Empty
            Else
                Select Case xlRow.GetCell(CurrentColumn).CellType
                    Case CellType.STRING
                        Return xlRow.GetCell(CurrentColumn).StringCellValue
                    Case CellType.NUMERIC
                        Return xlRow.GetCell(CurrentColumn).NumericCellValue
                    Case CellType.BOOLEAN
                        Return xlRow.GetCell(CurrentColumn).BooleanCellValue
                    Case CellType.FORMULA
                        Return xlRow.GetCell(CurrentColumn).StringCellValue 'force to string
                    Case Else
                        Return IIf(ReturnEmptyOnError, String.Empty, Nothing)
                End Select
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as Integer type. Strings are attempted to be converted to Integer
        ''' </summary>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or 0 if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCellAsInteger(Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Integer
            Dim int As Object = ReadCell(Column, Row)
            If int.GetType Is GetType(Integer) Then
                Return int
            ElseIf int.GetType Is GetType(String) Then
                Dim xlInteger As Integer
                Integer.TryParse(int, xlInteger)
                Return xlInteger
            Else
                Return 0
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as Double type. Strings are attempted to be converted to Double
        ''' </summary>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or 0 if read error</returns>
        ''' <remarks></remarks>
        Public Function ReadCellAsDouble(Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Double
            Dim dbl As Object = ReadCell(Column, Row)
            If dbl.GetType Is GetType(Double) Then
                Return dbl
            ElseIf dbl.GetType Is GetType(String) Then
                Dim xlDouble As Double
                Double.TryParse(dbl, xlDouble)
                Return xlDouble
            Else
                Return 0
            End If
        End Function

        ''' <summary>
        ''' Read the value at the next cell (or at the specified cell) as DateTime type. Strings are attempted to be converted to DateTime
        ''' </summary>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>Value or MinValue if read error</returns>
        ''' <remarks></remarks>   
        Public Function ReadCellAsDate(Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As DateTime
            Dim dbl As Object = ReadCell(Column, Row)
            If dbl.GetType Is GetType(Double) Then
                Return xlRow.GetCell(CurrentColumn).DateCellValue
            ElseIf dbl.GetType Is GetType(String) Then
                Dim xlDate As DateTime
                DateTime.TryParse(dbl, xlDate)
                Return xlDate
            Else
                Return DateTime.MinValue
            End If
        End Function

        ''' <summary>
        ''' Write the value at the next cell (or at the specified cell). 
        ''' </summary>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteCell(ByVal Value As Object, Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                    xlRow = xlSheet.GetRow(Row)
                    If IsNothing(xlRow) Then
                        xlRow = xlSheet.CreateRow(Row)
                        CurrentColumn = -1
                    End If
                Else
                    If IsNothing(xlRow) Then
                        If CurrentRow = -1 Then CurrentRow = 0
                        xlRow = xlSheet.CreateRow(CurrentRow)
                        CurrentColumn = -1
                    End If
                End If
                If Column > -1 Then
                    CurrentColumn = Column
                Else
                    CurrentColumn += 1
                End If
                If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then xlCell = xlRow.CreateCell(CurrentColumn)
                If Value.GetType Is GetType(String) Then
                    xlCell.SetCellType(CellType.STRING)
                ElseIf Value.GetType Is GetType(Integer) Or GetType(ValueType) Is GetType(Decimal) Or GetType(ValueType) Is GetType(Double) Then
                    xlCell.SetCellType(CellType.NUMERIC)
                ElseIf Value.GetType Is GetType(DateTime) Then
                    If Not IsDBNull(Value) Then
                        If Value.TimeOfDay.TotalMilliseconds = 0 Then
                            xlCell.CellStyle = CellStyleDate
                        Else
                            xlCell.CellStyle = CellStyleDateTime
                        End If
                        xlCell.SetCellValue(Value)
                    End If
                ElseIf Value.GetType Is GetType(Boolean) Then
                    xlCell.SetCellType(CellType.BOOLEAN)
                End If
                xlCell.SetCellValue(Value)
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Write the date/teime value at the next cell (or at the specified cell) with the specified format. 
        ''' </summary>
        ''' <param name="Value">Date/Time value</param>
        ''' <param name="DateTimeFormat">Format to display the date/time in. Nothing/Empty=default thread culture. ie: "dd/mm/yyyy hh:mm:ss AM/PM"</param>
        ''' <param name="Column">If provided, skip to the specified column. 0 based index (Excel iteself is 1 based index).</param>
        ''' <param name="Row">If provided, read from the specified row. 0 based index (Excel iteself is 1 based index).</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteCellAsDate(ByVal Value As DateTime, ByVal DateTimeFormat As String, Optional ByVal Column As Integer = -1, Optional ByVal Row As Integer = -1) As Boolean
            Try
                If Row > -1 Then
                    CurrentRow = Row
                    xlRow = xlSheet.GetRow(Row)
                    If IsNothing(xlRow) Then
                        xlRow = xlSheet.CreateRow(Row)
                        CurrentColumn = -1
                    End If
                Else
                    If IsNothing(xlRow) Then
                        If CurrentRow = -1 Then CurrentRow = 0
                        xlRow = xlSheet.CreateRow(CurrentRow)
                        CurrentColumn = -1
                    End If
                End If
                If Column > -1 Then
                    CurrentColumn = Column
                Else
                    CurrentColumn += 1
                End If
                If CurrentColumn > xlRow.PhysicalNumberOfCells - 1 Then xlCell = xlRow.CreateCell(CurrentColumn)
                If Not IsDBNull(Value) Then
                    If String.IsNullOrEmpty(DateTimeFormat) Then
                        If Value.TimeOfDay.TotalMilliseconds = 0 Then
                            xlCell.CellStyle = CellStyleDate
                        Else
                            xlCell.CellStyle = CellStyleDateTime
                        End If
                    Else
                        If Not CellStyleList.ContainsKey(DateTimeFormat) Then
                            Dim format As HSSFDataFormat = xlWorkbook.CreateDataFormat()
                            Dim xlStyle As HSSFCellStyle = xlWorkbook.CreateCellStyle()
                            xlStyle.DataFormat = format.GetFormat(DateTimeFormat)
                            CellStyleList.Add(DateTimeFormat, xlStyle)
                        End If
                        xlCell.CellStyle = CellStyleList(DateTimeFormat)
                    End If
                    xlCell.SetCellValue(Value)
                End If
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Set the cell style of the last cell
        ''' </summary>
        ''' <param name="Bold">Bold font. Default = False</param>
        ''' <param name="Align">Cell alignment. Default = NPOI.SS.UserModel.HorizontalAlignment.GENERAL</param>
        ''' <param name="Width">Cell Width. Default = 10</param>
        ''' <param name="CellSpan">Cell span. Default = 1</param>
        ''' <param name="FGColour">Font colour. none=System.Drawing.Color.Empty</param>
        ''' <param name="BGColour">Background colour. none=System.Drawing.Color.Empty</param>
        ''' <param name="Border">Thin black border. Default = False</param>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function FormatLastCell(ByVal Bold As Boolean, ByVal Align As HorizontalAlignment, ByVal Width As Integer, ByVal CellSpan As Integer, ByVal FGColour As System.Drawing.Color, ByVal BGColour As System.Drawing.Color, ByVal Border As Boolean) As Boolean
            Dim strStyleIndex As String = String.Concat(Bold.ToString, Align.ToString, FGColour.ToString, BGColour.ToString, Border.ToString)
            'reuse existing styles (avoid limit to number of new styles)
            If Not CellStyleList.ContainsKey(strStyleIndex) Then
                Dim xlStyle As HSSFCellStyle = xlCell.CellStyle
                If IsNothing(xlStyle) Then xlStyle = xlWorkbook.CreateCellStyle()
                If Bold Or FGColour <> System.Drawing.Color.Empty Then
                    Dim xlFont As HSSFFont = xlWorkbook.CreateFont()
                    If FGColour <> System.Drawing.Color.Empty Then xlFont.Color = GetXLColourFG(FGColour)
                    If Bold Then xlFont.Boldweight = FontBoldWeight.BOLD
                    xlStyle.SetFont(xlFont)
                End If
                If BGColour <> System.Drawing.Color.Empty Then
                    If BGColour <> System.Drawing.Color.Empty Then xlStyle.FillForegroundColor = GetXLColourBG(BGColour)
                    xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND
                End If
                If Border Then
                    xlStyle.BorderTop = CellBorderType.THIN
                    xlStyle.BorderLeft = CellBorderType.THIN
                    xlStyle.BorderRight = CellBorderType.THIN
                    xlStyle.BorderBottom = CellBorderType.THIN
                End If
                If Align <> HorizontalAlignment.GENERAL Then
                    xlStyle.Alignment = Align
                End If
                CellStyleList.Add(strStyleIndex, xlStyle)
            End If
            xlCell.CellStyle = CellStyleList(strStyleIndex)
            If CellSpan > 1 Then
                Dim LastMergedCells As New NPOI.SS.Util.CellRangeAddress(CurrentRow, CurrentRow, CurrentColumn, CurrentColumn + CellSpan)
                xlSheet.AddMergedRegion(LastMergedCells)
                If Border Then
                    'xlSheet.SetEnclosedBorderOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                    xlSheet.SetBorderTopOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                    xlSheet.SetBorderBottomOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                    xlSheet.SetBorderLeftOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                    xlSheet.SetBorderRightOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                End If
            End If
            xlSheet.SetColumnWidth(CurrentColumn, Width * 256)
        End Function

        ''' <summary>
        ''' Creates a new row for writing. Use this if using WriteCell with no row/column specified.
        ''' </summary>
        ''' <returns>True if ok</returns>
        ''' <remarks></remarks>
        Public Function WriteNewRow() As Boolean
            CurrentRow += 1
            Try
                xlRow = xlSheet.CreateRow(CurrentRow)
                CurrentColumn = -1
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Save the workbook to a new file
        ''' </summary>
        ''' <param name="XlsFilePath">Absolute path and filename for the .xls</param>
        ''' <returns>True if saved</returns>
        ''' <remarks></remarks>
        Public Function Save(ByVal XlsFilePath As String) As Boolean
            Try
                If File.Exists(XlsFilePath) Then File.Delete(XlsFilePath)
                Using fs As New FileStream(XlsFilePath, FileMode.Create)
                    xlWorkbook.Write(fs)
                    fs.Flush()
                    fs.Close()
                End Using
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function

        ''' <summary>
        ''' Save the workbook to a stream or filestream
        ''' </summary>
        ''' <param name="XlsFileStream">Stream or FileStream to save the .xls to</param>
        ''' <returns>True if saved</returns>
        ''' <remarks></remarks>
        Public Function Save(ByRef XlsFileStream As Stream) As Boolean
            Try
                xlWorkbook.Write(XlsFileStream)
                Return True
            Catch ex As Exception
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
                Return False
            End Try
        End Function


        ''' <summary>
        ''' Save a GridView Control to an Excel 97+ BIFF .xls file
        ''' </summary>
        ''' <param name="GridViewControl">Databound (DataBind already called) GridView Control</param>
        ''' <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
        ''' <returns>True if successful</returns>
        ''' <remarks></remarks>
        Public Function GridViewToXLS(ByRef GridViewControl As GridView, ByVal FileToSave As String) As Boolean
            Try
                If IO.File.Exists(FileToSave) Then IO.File.Delete(FileToSave)
            Catch
                'HttpContext.Current.Trace.Warn("Can not delete previous file")
            End Try
            If Not FileToSave.Contains(".") Then FileToSave &= ".xls"
            Dim fiSave As New FileInfo(FileToSave)
            Dim ReportName As String = fiSave.Name.Substring(0, fiSave.Name.IndexOf("."))
            ReportName = Regex.Replace(ReportName, "[^A-Z]", String.Empty, RegexOptions.IgnoreCase)
            Dim blnNoError As Boolean = True

            xlWorkbook = New HSSFWorkbook()
            CreateStandardProperties(ReportName)

            xlSheet.DisplayGridlines = False
            Dim xlStyle As HSSFCellStyle
            Dim xlFont As HSSFFont
            Dim aryWidths() As Int16
            ReDim aryWidths(0)
            CurrentRow = 0
            'xlWorkbook.NumCellStyles
            'xlWorkbook.GetCellStyleAt()
            Dim LastMergedCells As NPOI.SS.Util.CellRangeAddress
            Dim blnIsDate, blnIsDecimal As Boolean

            'Write tables
            Dim regExHTML As New Regex("<([^>]*)>[^<]*<\/\1>|<[^>]+/>|<[^>]+>", RegexOptions.Multiline Or RegexOptions.IgnoreCase)
            Dim TableCell As System.Web.UI.WebControls.TableCell

            'Header
            If Not IsNothing(GridViewControl.HeaderRow) Then
                xlRow = xlSheet.CreateRow(CurrentRow)
                xlStyle = xlWorkbook.CreateCellStyle()
                xlFont = xlWorkbook.CreateFont()
                xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index
                xlFont.Boldweight = FontBoldWeight.BOLD
                xlStyle.SetFont(xlFont)
                xlStyle.FillForegroundColor = GetXLColourBG(System.Drawing.ColorTranslator.FromHtml("#4666ad"))
                xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND
                ReDim aryWidths(GridViewControl.HeaderRow.Cells.Count - 1)
                For Each TableCell In GridViewControl.HeaderRow.Cells
                    If GridViewControl.Columns(CurrentColumn).Visible Then
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(Regex.Replace(TableCell.Text, "<[^>]*>", " ")))
                        xlCell.CellStyle = xlStyle
                        If TableCell.Text.Length + 1 > aryWidths(CurrentColumn) Then
                            aryWidths(CurrentColumn) = TableCell.Text.Length + 1
                        End If
                        CurrentColumn += 1
                    End If
                Next
                CurrentRow += 1
            End If

            'Data rows
            Dim intCol, intRow, intValue, intCellSize, ActualCol As Integer
            Dim decValue As Decimal
            Dim strValue As String
            Dim GridViewRowControl As GridViewRow
            For intRow = 0 To GridViewControl.Rows.Count - 1
                GridViewRowControl = GridViewControl.Rows(intRow)
                xlRow = xlSheet.CreateRow(CurrentRow)
                If aryWidths.GetUpperBound(0) < GridViewRowControl.Cells.Count Then ReDim Preserve aryWidths(GridViewRowControl.Cells.Count - 1)
                ActualCol = 0
                For intCol = 0 To GridViewRowControl.Cells.Count - 1
                    TableCell = GridViewRowControl.Cells(intCol)
                    If GridViewControl.Columns(intCol).Visible Then
                        If TableCell.Controls.Count > 0 Then
                            strValue = String.Empty
                            For Each cntl As Control In TableCell.Controls
                                If cntl.GetType Is GetType(DataBoundLiteralControl) Then
                                    strValue &= DirectCast(cntl, DataBoundLiteralControl).Text.Trim
                                ElseIf cntl.GetType Is GetType(CheckBox) Then
                                    strValue &= DirectCast(cntl, CheckBox).Checked.ToString
                                ElseIf cntl.GetType Is GetType(SymTextBox) Then
                                    strValue &= DirectCast(cntl, SymTextBox).Text
                                ElseIf cntl.GetType Is GetType(DropDownList) Then
                                    strValue &= DirectCast(cntl, DropDownList).SelectedItem.Text
                                ElseIf cntl.GetType Is GetType(ListBox) Then
                                    For Each li As ListItem In DirectCast(cntl, ListBox).Items
                                        If li.Selected Then
                                            If strValue <> String.Empty Then strValue &= ", "
                                            strValue &= li.Text
                                        End If
                                    Next
                                Else
                                    strValue &= TableCell.Text
                                End If
                            Next
                        Else
                            strValue = TableCell.Text
                        End If

                        'Merge/span cells
                        If TableCell.ColumnSpan > 1 Or TableCell.RowSpan > 1 Then
                            LastMergedCells = New NPOI.SS.Util.CellRangeAddress(CurrentRow, CurrentRow, ActualCol, ActualCol + TableCell.ColumnSpan - 1)
                            xlSheet.AddMergedRegion(LastMergedCells)
                        Else
                            LastMergedCells = Nothing
                        End If
                        'xlCell = xlRow.Cells.Add()
                        xlCell = xlRow.CreateCell(ActualCol)

                        'reuse existing styles (avoid limit to number of new styles)
                        blnIsDate = IsDate(strValue)
                        blnIsDecimal = strValue.Contains(".") AndAlso Decimal.TryParse(strValue, System.Globalization.NumberStyles.Currency, Nothing, decValue)
                        If blnIsDate Or blnIsDecimal Or TableCell.Font.Bold Or TableCell.ForeColor <> Color.Empty Or TableCell.BackColor <> Color.Empty Or GridViewControl.GridLines > GridLines.None Or TableCell.BorderWidth.Value > 0 Or _
                                               TableCell.HorizontalAlign <> HorizontalAlign.NotSet Then
                            Dim strStyle As String = String.Concat(blnIsDate.ToString, IIf(blnIsDate, TableCell.Text.Contains(":").ToString, String.Empty), blnIsDecimal.ToString, _
                               TableCell.Font.Bold.ToString, TableCell.ForeColor, TableCell.BackColor, _
                               GridViewControl.GridLines.ToString, TableCell.BorderWidth.Value.ToString, TableCell.HorizontalAlign.ToString)

                            If Not IsNothing(LastMergedCells) AndAlso (GridViewControl.GridLines > GridLines.None Or TableCell.BorderWidth.Value > 0) Then
                                'xlSheet.SetEnclosedBorderOfRegion(LastMergedCells, CellBorderType.DASHED, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                                xlSheet.SetBorderTopOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                                xlSheet.SetBorderBottomOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                                xlSheet.SetBorderLeftOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                                xlSheet.SetBorderRightOfRegion(LastMergedCells, CellBorderType.THIN, NPOI.HSSF.Util.HSSFColor.BLACK.index)
                            End If

                            If Not CellStyleList.ContainsKey(strStyle) Then
                                xlStyle = xlWorkbook.CreateCellStyle()
                                'Colour
                                If TableCell.Font.Bold Or TableCell.ForeColor <> Color.Empty Then
                                    xlFont = xlWorkbook.CreateFont()
                                    If TableCell.ForeColor <> Color.Empty Then xlFont.Color = GetXLColourFG(TableCell.ForeColor)
                                    If TableCell.Font.Bold Then xlFont.Boldweight = FontBoldWeight.BOLD
                                    xlStyle.SetFont(xlFont)
                                End If
                                If TableCell.BackColor <> Color.Empty Then
                                    xlStyle.FillForegroundColor = GetXLColourBG(TableCell.BackColor)
                                    xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND
                                End If

                                If GridViewControl.GridLines > GridLines.None Or TableCell.BorderWidth.Value > 0 Then
                                    'Border/Grid
                                    xlStyle.BorderTop = CellBorderType.THIN
                                    xlStyle.BorderLeft = CellBorderType.THIN
                                    xlStyle.BorderRight = CellBorderType.THIN
                                    xlStyle.BorderBottom = CellBorderType.THIN
                                End If

                                'Alignment
                                If TableCell.HorizontalAlign <> HorizontalAlign.NotSet Then
                                    Select Case TableCell.HorizontalAlign
                                        Case HorizontalAlign.Left
                                            xlStyle.Alignment = HorizontalAlignment.LEFT
                                        Case HorizontalAlign.Right
                                            xlStyle.Alignment = HorizontalAlignment.RIGHT
                                        Case HorizontalAlign.Center
                                            xlStyle.Alignment = HorizontalAlignment.CENTER
                                    End Select
                                End If

                                If blnIsDate Then
                                    'Date format
                                    If strValue.Contains(":") Then
                                        xlStyle.DataFormat = FormatDateTime
                                        intCellSize = DateTimeFormat.Length
                                    Else
                                        xlStyle.DataFormat = FormatDate
                                        intCellSize = DateFormat.Length
                                    End If
                                ElseIf blnIsDecimal Then
                                    'Common decimal styles
                                    If TableCell.Text.Length - TableCell.Text.IndexOf(".") = 3 Then
                                        xlStyle.DataFormat = HSSFDataFormat.GetBuiltinFormat("0.00")
                                    ElseIf TableCell.Text.Length - TableCell.Text.IndexOf(".") = 2 Then
                                        xlStyle.DataFormat = HSSFDataFormat.GetBuiltinFormat("0.0")
                                    End If
                                End If

                                CellStyleList.Add(strStyle, xlStyle)
                            End If
                            xlCell.CellStyle = CellStyleList(strStyle)
                        End If

                        If Integer.TryParse(strValue, System.Globalization.NumberStyles.Integer Or System.Globalization.NumberStyles.AllowThousands, Nothing, intValue) Then
                            xlCell.SetCellType(CellType.NUMERIC)
                            xlCell.SetCellValue(intValue)
                            intCellSize = strValue.Length
                        ElseIf blnIsDecimal Then
                            xlCell.SetCellType(CellType.NUMERIC)
                            xlCell.SetCellValue(decValue)
                            intCellSize = strValue.Length
                        ElseIf blnIsDate Then
                            xlCell.SetCellValue(strValue)
                        ElseIf regExHTML.IsMatch(strValue) Then
                            strValue = System.Web.HttpUtility.HtmlDecode(Regex.Replace(strValue, "<[^>]*>", String.Empty))
                            If Decimal.TryParse(strValue, decValue) Then
                                xlCell.SetCellType(CellType.NUMERIC)
                                xlCell.SetCellValue(decValue)
                            Else
                                xlCell.SetCellValue(strValue)
                            End If
                            intCellSize = strValue.Length
                        Else
                            xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(strValue))
                            intCellSize = xlCell.StringCellValue.Length
                        End If

                        If intCellSize > aryWidths(intCol) AndAlso GridViewRowControl.Cells(intCol).ColumnSpan < 2 Then
                            aryWidths(intCol) = intCellSize
                        End If
                        ActualCol += IIf(TableCell.ColumnSpan = 0, 1, TableCell.ColumnSpan)
                    End If
                Next
                CurrentRow += 1
            Next

            'Footer
            If Not IsNothing(GridViewControl.FooterRow) Then
                xlRow = xlSheet.CreateRow(CurrentRow)
                xlStyle = xlWorkbook.CreateCellStyle()
                xlFont = xlWorkbook.CreateFont()
                xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index
                xlFont.Boldweight = FontBoldWeight.BOLD
                xlStyle.SetFont(xlFont)
                xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.GREY_50_PERCENT.index
                xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND
                CurrentColumn = 0
                For Each TableCell In GridViewControl.FooterRow.Cells
                    If GridViewControl.Columns(CurrentColumn).Visible Then
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlCell.SetCellValue(System.Web.HttpUtility.HtmlDecode(Regex.Replace(TableCell.Text, "<[^>]*>", String.Empty)))
                        xlStyle.Alignment = HorizontalAlignment.RIGHT
                        xlCell.CellStyle = xlStyle
                        If TableCell.Text.Length + 1 > aryWidths(CurrentColumn) Then
                            aryWidths(CurrentColumn) = TableCell.Text.Length + 1
                        End If
                        CurrentColumn += 1
                    End If
                Next
            End If

            'Set column widths (NPOI auto-width can be ugly)
            For intColumn As Integer = 0 To aryWidths.GetUpperBound(0)
                If aryWidths(intColumn) > 50 Then
                    xlSheet.SetColumnWidth(intColumn, 50 * 256)
                ElseIf aryWidths(intColumn) < 10 Then
                    xlSheet.SetColumnWidth(intColumn, 10 * 256)
                Else
                    xlSheet.SetColumnWidth(intColumn, (aryWidths(intColumn) + 1) * 256)
                End If
            Next

            Try
                If FileToSave.IndexOf(":") < 1 Then
                    'just filename to stream to browser and force it to download the file (and launch Excel or save, not display in the browser)
                    Dim ms As New MemoryStream()
                    xlWorkbook.Write(ms)
                    HttpContext.Current.Response.Clear()
                    HttpContext.Current.Response.AddHeader("Content-Disposition", "attachment; filename=" & FileToSave)
                    HttpContext.Current.Response.AddHeader("Content-Length", ms.Length.ToString())
                    HttpContext.Current.Response.ContentType = "application/octet-stream"
                    HttpContext.Current.Response.BinaryWrite(ms.GetBuffer())
                    HttpContext.Current.Response.Flush()
                    HttpContext.Current.Response.End()
                Else
                    Using fs As New FileStream(FileToSave, FileMode.Create)
                        xlWorkbook.Write(fs)
                        fs.Close()
                    End Using
                End If
            Catch ex As Exception
                blnNoError = False
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
            End Try
            Return blnNoError
        End Function

        ''' <summary>
        ''' Save a Table to an Excel 97+ BIFF .xls file
        ''' </summary>
        ''' <param name="SourceData">Source table</param>
        ''' <param name="SourceColOrder">Source column name order. Nothing if auto order (per table). Required if SourceColText specified</param>
        ''' <param name="SourceColText">Source column header text. Nothing if auto names (per table). Required if SourceColOrder specified</param>
        ''' <param name="FileToSave">File Name with absolute path to save to file, or just filename for immediate stream to Response</param>
        ''' <param name="IncludeHeader">Write column names</param>
        ''' <returns>True if successful</returns>
        ''' <remarks></remarks>
        Public Overloads Function TableToXLS(ByRef SourceData As DataTable, ByRef SourceColOrder() As String, ByRef SourceColText() As String, ByVal IncludeHeader As Boolean, ByVal FileToSave As String) As Boolean
            Try
                If IO.File.Exists(FileToSave) Then IO.File.Delete(FileToSave)
            Catch
                'HttpContext.Current.Trace.Warn("Can not delete previous file")
            End Try
            Dim fiSave As New FileInfo(FileToSave)
            Dim SheetName As String = fiSave.Name.Substring(0, fiSave.Name.IndexOf("."))
            SheetName = Regex.Replace(SheetName, "[^A-Z]", String.Empty, RegexOptions.IgnoreCase)
            Dim blnNoError As Boolean = True

            xlWorkbook = New HSSFWorkbook()
            CreateStandardProperties(SheetName)

            xlSheet = xlWorkbook.CreateSheet(SheetName)
            xlSheet.DisplayGridlines = True
            Dim xlStyle As HSSFCellStyle
            Dim aryWidths() As Int16
            ReDim aryWidths(SourceData.Columns.Count - 1)
            CurrentRow = 0

            'Write header
            If IncludeHeader Then
                xlRow = xlSheet.CreateRow(CurrentRow)
                xlStyle = xlWorkbook.CreateCellStyle()
                Dim xlFont As HSSFFont = xlWorkbook.CreateFont()
                xlFont.Color = NPOI.HSSF.Util.HSSFColor.WHITE.index

                xlFont.Boldweight = FontBoldWeight.BOLD
                xlStyle.SetFont(xlFont)
                xlStyle.FillForegroundColor = NPOI.HSSF.Util.HSSFColor.ROYAL_BLUE.index
                xlStyle.FillPattern = FillPatternType.SOLID_FOREGROUND

                If IsNothing(SourceColText) Then
                    'Auto names
                    ReDim SourceColOrder(SourceData.Columns.Count - 1)
                    For Each dcCol As DataColumn In SourceData.Columns
                        SourceColOrder(CurrentColumn) = dcCol.ColumnName
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlCell.SetCellValue(dcCol.ColumnName.Replace("_", " "))
                        xlCell.CellStyle = xlStyle
                        If dcCol.ColumnName.Length + 1 > aryWidths(CurrentColumn) Then
                            aryWidths(CurrentColumn) = dcCol.ColumnName.Length + 1
                        End If
                        CurrentColumn += 1
                    Next
                Else
                    'Specified names
                    For Each strName As String In SourceColText
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        xlCell.SetCellValue(strName)
                        xlCell.CellStyle = xlStyle
                        If strName.Length + 1 > aryWidths(CurrentColumn) Then
                            aryWidths(CurrentColumn) = strName.Length + 1
                        End If
                        CurrentColumn += 1
                    Next
                End If
            End If

            If IsNothing(SourceColOrder) Then
                'Get column names if not specified
                CurrentColumn = 0
                ReDim SourceColOrder(SourceData.Columns.Count - 1)
                For Each dcCol As DataColumn In SourceData.Columns
                    SourceColOrder(CurrentColumn) = dcCol.ColumnName
                    CurrentColumn += 1
                Next
            End If

            'Write data rows
            Dim SourceDataCell As Object
            For Each drvRow As DataRow In SourceData.Rows
                CurrentRow += 1
                CurrentColumn = 0
                xlRow = xlSheet.CreateRow(CurrentRow)
                For Each strColName As String In SourceColOrder 'For CurrentColumn = 0 To SourceData.Columns.Count - 1
                    If strColName <> String.Empty Then
                        SourceDataCell = drvRow(strColName)
                        xlCell = xlRow.CreateCell(CurrentColumn)
                        If SourceDataCell.GetType Is GetType(System.Int32) Or SourceDataCell.GetType Is GetType(System.Int64) Then
                            xlCell.SetCellType(CellType.NUMERIC)
                            If Not IsDBNull(SourceDataCell) Then xlCell.SetCellValue(SourceDataCell)
                        ElseIf SourceDataCell.GetType Is GetType(System.DateTime) Then
                            If Not IsDBNull(SourceDataCell) Then
                                If SourceDataCell.TimeOfDay.TotalMilliseconds = 0 Then
                                    xlCell.CellStyle = CellStyleDate
                                Else
                                    xlCell.CellStyle = CellStyleDateTime
                                End If
                                xlCell.SetCellValue(SourceDataCell)
                            End If
                        ElseIf SourceDataCell.GetType Is GetType(System.Double) Or SourceDataCell.GetType Is GetType(System.Decimal) Then
                            xlCell.SetCellType(CellType.NUMERIC)
                            If Not IsDBNull(SourceDataCell) Then xlCell.SetCellValue(SourceDataCell)
                        Else
                            xlCell.SetCellValue(SourceDataCell.ToString)
                        End If
                        If SourceDataCell.ToString.Length > aryWidths(CurrentColumn) Then
                            aryWidths(CurrentColumn) = SourceDataCell.ToString.Length
                        End If
                        CurrentColumn += 1
                    End If
                Next
            Next

            'Set column widths (NPOI auto-width can be ugly)
            For intColumn As Integer = 0 To aryWidths.GetUpperBound(0)
                If aryWidths(intColumn) > 50 Then
                    xlSheet.SetColumnWidth(intColumn, 50 * 256)
                ElseIf aryWidths(intColumn) < 10 Then
                    xlSheet.SetColumnWidth(intColumn, 10 * 256)
                Else
                    xlSheet.SetColumnWidth(intColumn, (aryWidths(intColumn) + 1) * 256)
                End If
            Next

            Try
                If FileToSave.IndexOf(":") < 1 Then
                    'just filename to stream to browser and force it to download the file (and launch Excel or save, not display in the browser)
                    Dim ms As New MemoryStream()
                    xlWorkbook.Write(ms)
                    HttpContext.Current.Response.Clear()
                    HttpContext.Current.Response.AddHeader("Content-Disposition", "attachment; filename=" & FileToSave)
                    HttpContext.Current.Response.AddHeader("Content-Length", ms.Length.ToString())
                    HttpContext.Current.Response.ContentType = "application/octet-stream"
                    HttpContext.Current.Response.BinaryWrite(ms.GetBuffer())
                    HttpContext.Current.Response.Flush()
                    HttpContext.Current.Response.End()
                Else
                    Using fs As New FileStream(FileToSave, FileMode.Create)
                        xlWorkbook.Write(fs)
                        fs.Close()
                    End Using
                End If
            Catch ex As Exception
                blnNoError = False
                ErrorMessage = ex.Message
                'HttpContext.Current.Trace.Warn(ex.ToString)
            End Try
            Return blnNoError
        End Function

        Private Sub CreateDateStyles()
            'Create a default date and date/time format and style for general use
            Dim format As HSSFDataFormat = xlWorkbook.CreateDataFormat()
            DateFormat = Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortDatePattern.ToLower
            FormatDate = format.GetFormat(DateFormat)
            format = xlWorkbook.CreateDataFormat()
            CellStyleDate = xlWorkbook.CreateCellStyle()
            CellStyleDate.DataFormat = FormatDate
            DateTimeFormat = String.Concat(DateFormat, " ", Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.ShortTimePattern.ToLower.Replace("tt", "AM/PM"))
            FormatDateTime = format.GetFormat(DateTimeFormat)
            CellStyleDateTime = xlWorkbook.CreateCellStyle()
            CellStyleDateTime.DataFormat = FormatDateTime
        End Sub

        Private Function GetXLColourFG(ByVal SystemColour As System.Drawing.Color) As Short
            'Lookup RGB from .NET system colour in Excel pallete - or nearest match. Font colour in 1.2.3 has a bug and does not support custom entries
            If SystemColour = System.Drawing.Color.Empty Then Return NPOI.HSSF.Util.HSSFColor.COLOR_NORMAL 'COLOR_NORMAL=unspecified. 'alt= AUTOMATIC.index 
            Dim XlPalette As HSSFPalette = xlWorkbook.GetCustomPalette()
            Dim XlColour As NPOI.HSSF.Util.HSSFColor = XlPalette.FindColor(SystemColour.R, SystemColour.G, SystemColour.B)
            If IsNothing(XlColour) Then
                XlColour = XlPalette.FindSimilarColor(SystemColour.R, SystemColour.G, SystemColour.B)
                Return XlColour.GetIndex()
            Else
                Return XlColour.GetIndex()
            End If
        End Function

        Private Function GetXLColourBG(ByVal SystemColour As System.Drawing.Color) As Short
            'Lookup RGB from .NET system colour in Excel pallete - or create a new entry (get nearest if palette full). Return the XL palette index.
            If SystemColour = System.Drawing.Color.Empty Then Return NPOI.HSSF.Util.HSSFColor.COLOR_NORMAL 'COLOR_NORMAL=unspecified. 'alt= AUTOMATIC.index 
            Dim XlPalette As HSSFPalette = xlWorkbook.GetCustomPalette()
            Dim XlColour As NPOI.HSSF.Util.HSSFColor = XlPalette.FindColor(SystemColour.R, SystemColour.G, SystemColour.B)
            If IsNothing(XlColour) Then
                'Available colour palette entries: 65 to 32766 (0-64=standard palette; 64=auto, 32767=unspecified)
                If NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE < 255 Then
                    If NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE < 67 Then
                        'Move start pointer up. First Entry seems to fail, so add dummy.
                        NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE = 67
                        XlColour = XlPalette.AddColor(100, 100, 100)
                        XlColour = XlPalette.AddColor(100, 100, 100)
                    End If
                    NPOI.HSSF.Record.PaletteRecord.STANDARD_PALETTE_SIZE += 1
                    XlColour = XlPalette.AddColor(SystemColour.R, SystemColour.G, SystemColour.B)
                Else
                    XlColour = XlPalette.FindSimilarColor(SystemColour.R, SystemColour.G, SystemColour.B)
                End If
                Return XlColour.GetIndex()
            Else
                Return XlColour.GetIndex()
            End If
        End Function

    End Class

'End Namespace